from ..gen_platonic.mid_pt import mid_pt
from .dode_face_COM_coor import dode_face_COM_coor


def dode_face_COM_leg_coor(a: float, b: float, c: float, d: float, e: float):
    """Calculates the center of mass (COM) and the coordinates of the five legs
    of a protein based on five input coordinates on the same face of a dodecahedron.

    Args:
        a (float): The first coordinate on the face.
        b (float): The second coordinate on the face.
        c (float): The third coordinate on the face.
        d (float): The fourth coordinate on the face.
        e (float): The fifth coordinate on the face.

    Returns:
        list: A list of six lists, where the first element is a list of three
        float values representing the X, Y, and Z coordinates of the center of mass (COM),
        and the remaining five elements represent the coordinates of the five legs of
        the protein, calculated as midpoints between the input coordinates.

    Raises:
        None.

    Example:
        >>> dode_face_COM_leg_coor([0.0, 0.0, 0.0], [1.0, 1.0, 1.0],
        ...                        [2.0, 2.0, 2.0], [3.0, 3.0, 3.0], [4.0, 4.0, 4.0])
        [[0.29389262614624, 0.29389262614624, 0.29389262614624],
         [0.5, 0.5, 0.5],
         [1.5, 1.5, 1.5],
         [2.5, 2.5, 2.5],
         [3.5, 3.5, 3.5],
         [4.0, 4.0, 4.0]]

    Note:
        - The function returns a list of six lists, where the first element is the
          COM coordinates and the remaining five elements represent the coordinates
          of the five legs of the protein.
    """
    
    # calculate COM and 5 legs of one protein, 6 coords in total [COM, lg1, lg2, lg3, lg4, lg5]
    COM_leg = []
    COM_leg.append(dode_face_COM_coor(a, b, c, d, e))
    COM_leg.append(mid_pt(a, b))
    COM_leg.append(mid_pt(b, c))
    COM_leg.append(mid_pt(c, d))
    COM_leg.append(mid_pt(d, e))
    COM_leg.append(mid_pt(e, a))
    return COM_leg





import numpy as np
from ..gen_platonic.mid_pt import mid_pt


def tetr_vert_COM_leg(COM: float, a: float, b: float, c: float):
    """Calculates the center of mass (COM) and midpoints of three edges of a tetrahedron.
    
    Args:
        COM (float): The center of mass of the tetrahedron.
        a (float): The first vertex of the tetrahedron.
        b (float): The second vertex of the tetrahedron.
        c (float): The third vertex of the tetrahedron.
        
    Returns:
        list: A list of four values, [COM, lega, legb, legc], rounded to 10 decimal places.
        
    Example:
        >>> tetr_vert_COM_leg(0.5, 1.0, 2.0, 3.0)
        [0.5, 0.75, 1.5, 2.25]
    """

    lega = mid_pt(COM, a)
    legb = mid_pt(COM, b)
    legc = mid_pt(COM, c)
    return [np.around(COM, 10), np.around(lega, 10), np.around(legb, 10), np.around(legc, 10)]





from .calculate_distance import calculate_distance

def determine_bind(site_1, site_2, buffer_ratio, site_dict, sigma):
    """Determines whether two sites are binded based on how close they are

    Args:
        site_1 (list): information about 1 site (import part: it's location)
        site_2 (list): information about another site
        buffer_ratio (float): the buffer ratio used to determine binding, defined as a fraction of 'sigma'
        site_dict (dictionary): turns column name (from df) into list index for site_array
        sigma (float): how close the two sites need to be in order to be considered binded

    Returns:
        boolean: whether or not they are bonded
    """
    #calculate distance
    x = (site_1[site_dict['x_coord']],site_1[site_dict['y_coord']],site_1[site_dict['z_coord']])
    y = (site_2[site_dict['x_coord']],site_2[site_dict['y_coord']],site_2[site_dict['z_coord']])
    distance = calculate_distance(x, y)

    #calculate if they are close enough to bind
    return distance >= sigma*(1-buffer_ratio) and distance <= sigma*(1+buffer_ratio)





from .octa_face_write import octa_face_write


def octa_face(radius: float, sigma: float):
    """Generate an octagonal face image.

    Args:
        radius (float): The radius of the octagonal face.
        sigma (float): The sigma value for generating the face.

    Returns:
        parm.inp/icos.mol: input files for NERDSS

    Raises:
        ValueError: If radius or sigma are not valid (e.g., negative values).

    Example:
        To generate an octagonal face with a radius of 10 and a sigma of 1.5:
        >>> octa_face(10, 1.5)
        File writing complete!
    """
    octa_face_write(radius, sigma)
    print('File writing complete!')
    return 0





import pandas as pd
import os
from .hist_to_csv import hist_to_csv


def hist_to_df(FullHist: list, SpeciesList: list, OpName: str, SaveCsv: bool = True, Single: bool = False):
    """Creates a pandas dataframe from a histogram.dat (multi-species)

    Args:
        FullHist (list): holds all of the histogram data
        SpeciesList (list): list of each included species type
        OpName (str): name of the outputted .csv file
        SaveCsv (bool, optional): If a .csv file is saved as well. Defaults to True.

    Returns:
       pandas.df: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
        is the number of that complex molecule at that time stamp.
    """
    hist_to_csv(FullHist, SpeciesList, OpName, Single)
    
    
    df = pd.read_csv(f'{OpName}.csv')
    if not SaveCsv:
        os.remove(f'{OpName}.csv')
    return df





import math
from ..gen_platonic.mid_pt import mid_pt


def icos_vert_center_coor(a: float, b: float, c: float, d: float, e: float):
    """Calculate the coordinates of the center of mass for an icosahedron.

    This function calculates the coordinates of the center of mass (COM) for
    an icosahedron, given the coordinates of five points (a, b, c, d, e) and
    using the mid_pt function from the ..gen_platonic.mid_pt module. The COM
    coordinates are computed based on the formula:
    COM = point + (mid_point - point) / (1 + sin(0.3 * pi))

    Args:
        a (float): The coordinates of point a as a list or tuple of three float values.
        b (float): The coordinates of point b as a list or tuple of three float values.
        c (float): The coordinates of point c as a list or tuple of three float values.
        d (float): The coordinates of point d as a list or tuple of three float values.
        e (float): The coordinates of point e as a list or tuple of three float values.

    Returns:
        list: The coordinates of the center of mass (COM) as a list of three float values.

    Example:
        >>> a = [1.0, 2.0, 3.0]
        >>> b = [4.0, 5.0, 6.0]
        >>> c = [7.0, 8.0, 9.0]
        >>> d = [10.0, 11.0, 12.0]
        >>> e = [13.0, 14.0, 15.0]
        >>> icos_vert_center_coor(a, b, c, d, e)
        [5.18101203220144, 6.58101203220144, 7.98101203220144]
    """
    n = 8
    mid_a = mid_pt(c, d)
    mid_b = mid_pt(d, e)
    mid_c = mid_pt(a, e)
    COM_a = []
    COM_b = []
    COM_c = []
    for i in range(0, 3):
        COM_a.append(round(a[i] + (mid_a[i] - a[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
        COM_b.append(round(b[i] + (mid_b[i] - b[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
        COM_c.append(round(c[i] + (mid_c[i] - c[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
    if round(COM_a[0], n) == round(COM_b[0], n) and round(COM_b[0], n) == round(COM_c[0], n) and \
        round(COM_a[1], n) == round(COM_b[1], n) and round(COM_b[1], n) == round(COM_c[1], n) and \
            round(COM_a[2], n) == round(COM_b[2], n) and round(COM_b[2], n) == round(COM_c[2], n):
        return COM_a
    else:
        return COM_a





import numpy as np


def real_PDB_norm_check(norm, COM, site, buffer_ratio=1e-3):
    """Check if the given 3D vector norm is normal to the plane defined by points COM and site in a real PDB file.

    Args:
        norm (List[float]): A 3D vector representing the normal to the plane.
        COM (List[float]): A 3D point representing the center of mass of the plane.
        site (List[float]): A 3D point representing a site in the plane.
        buffer_ratio (float, optional): A ratio used to define a buffer zone around the expected value of the norm's projection. Defaults to 1e-3.

    Returns:
        bool: True if the norm is normal to the plane within the given buffer zone, False otherwise.

    The function first checks the input lists for type and length correctness, and returns True if any error is found. Then, it computes the projection of the norm onto the plane, and checks whether it is within the expected range defined by the buffer ratio. The function returns True if the projection is within the range, and False otherwise.

    """
    for i in norm:
        if type(i) != float:
            return True
    for i in COM:
        if type(i) != float:
            return True
    for i in site:
        if type(i) != float:
            return True
    if len(norm) != 3 or len(COM) != 3 or len(site) != 3:
        return True
    if norm == [0, 0, 0]:
        return True
    
    norm = np.array(norm)
    COM = np.array(COM)
    site = np.array(site)
    vec1 = norm
    vec2 = site - COM
    zero_pos_1 = []
    zero_pos_2 = []
    for i in range(len(vec1)):
        if vec1[i] == 0:
            zero_pos_1.append(i)
    for i in range(len(vec2)):
        if vec2[i] == 0:
            zero_pos_2.append(i)
    
    
    if len(zero_pos_1) == 1 and len(zero_pos_2) == 1 and zero_pos_1 == zero_pos_2:
        pool = [0, 1, 2]
        pool.remove(zero_pos_1[0])
        ratio = vec1[pool[0]]/vec2[pool[0]]
        if vec1[pool[1]]/vec2[pool[1]] >= ratio*(1-buffer_ratio) and vec1[pool[1]]/vec2[pool[1]] <= ratio*(1+buffer_ratio):
            return True
        else:
            return False
    elif len(zero_pos_1) == 1 and len(zero_pos_2) == 1 and zero_pos_1 != zero_pos_2:
        return False
    elif len(zero_pos_1) == 2 and len(zero_pos_2) == 2 and zero_pos_1 == zero_pos_2:
        return True
    elif len(zero_pos_1) == 2 and len(zero_pos_2) == 2 and zero_pos_1 != zero_pos_2:
        return False
    elif len(zero_pos_1) != len(zero_pos_2):
        return False
    else:
        ratio = vec1[0]/vec2[0]
        if ratio >= 0:
            if vec1[1]/vec2[1] >= ratio*(1-buffer_ratio) and vec1[1]/vec2[1] <= ratio*(1+buffer_ratio):
                if vec1[2]/vec2[2] >= ratio*(1-buffer_ratio) and vec1[2]/vec2[2] <= ratio*(1+buffer_ratio):
                    return True
                else:
                    return False
            else:
                return False
        if ratio < 0:
            if vec1[1]/vec2[1] >= ratio*(1+buffer_ratio) and vec1[1]/vec2[1] <= ratio*(1-buffer_ratio):
                if vec1[2]/vec2[2] >= ratio*(1+buffer_ratio) and vec1[2]/vec2[2] <= ratio*(1-buffer_ratio):
                    return True
                else:
                    return False
            else:
                return False





import math
import numpy as np
from .real_PDB_unit import real_PDB_unit
from .real_PDB_triangle_correction import real_PDB_triangle_correction
from .real_PDB_calculate_phi import real_PDB_calculate_phi


def real_PDB_angles(COM1, COM2, int_site1, int_site2, normal_point1, normal_point2):
    """Calculates the angles for binding using the provided inputs.

    Args:
        COM1 (array): The center of mass of particle 1.
        COM2 (array): The center of mass of particle 2.
        int_site1 (array): The interface site of particle 1.
        int_site2 (array): The interface site of particle 2.
        normal_point1 (array): The normal point of particle 1.
        normal_point2 (array): The normal point of particle 2.

    Returns:
        Tuple of floats: The calculated values for theta1, theta2, phi1, phi2, omega, and sigma_magnitude respectively.

    Raises:
        ValueError: If n1 and v1 or n2 and v2 are parallel.
        """

    # Convert sequences into arrays for convinience
    COM1 = np.array(COM1)
    COM2 = np.array(COM2)
    int_site1 = np.array(int_site1)
    int_site2 = np.array(int_site2)
    normal_point1 = np.array(normal_point1)
    normal_point2 = np.array(normal_point2)

    # Get Vectors
    v1 = int_site1 - COM1  # from COM to interface (particle 1)
    v2 = int_site2 - COM2  # from COM to interface (particle 2)
    sigma1 = int_site1 - int_site2  # sigma, from p2 to p1
    sigma2 = int_site2 - int_site1  # sigma, from p1 to p2
    n1 = real_PDB_unit(normal_point1 - COM1)  # normal vector for p1
    n2 = real_PDB_unit(normal_point2 - COM2)  # normal vector for p2

    # Calculate the magnititude of sigma
    sigma_magnitude = np.linalg.norm(sigma1)

    # Calculate theta1 and theta2
    costheta1 = np.dot(v1, sigma1) / np.linalg.norm(v1) / \
        np.linalg.norm(sigma1)
    costheta2 = np.dot(v2, sigma2) / np.linalg.norm(v2) / \
        np.linalg.norm(sigma2)
    theta1 = math.acos(real_PDB_triangle_correction(costheta1))
    theta2 = math.acos(real_PDB_triangle_correction(costheta2))

    # check geometry
    errormsg = ''
    iferror = False  # determine if v // n
    if np.linalg.norm(np.cross(n1, v1)) < 10**-6:
        iferror = True
        errormsg += '\n\tn1 and v1 parallel, phi1 not available'
    if np.linalg.norm(np.cross(n2, v2)) < 10**-6:
        iferror = True
        errormsg += '\n\tn2 and v2 parallel, phi2 not available'
    if iferror:
        raise ValueError(errormsg)

    # determine if phi1 exists (v1 // sigma1 ?)
    if np.linalg.norm(np.cross(sigma1, v1)) < 10**-6:
        phi1 = float('nan')
        # omega_parallel = True
        omega_t1 = real_PDB_unit(np.cross(sigma1, n1))
    else:
        phi1 = real_PDB_calculate_phi(v1, n1, sigma1)
        omega_t1 = real_PDB_unit(np.cross(sigma1, v1))

    # determine if phi2 exists (v2 // sigma2 ?)
    if np.linalg.norm(np.cross(sigma2, v2)) < 10**-6:
        phi2 = float('nan')
        # omega_parallel = True
        omega_t2 = real_PDB_unit(np.cross(sigma1, n2))
    else:
        phi2 = real_PDB_calculate_phi(v2, n2, sigma2)
        omega_t2 = real_PDB_unit(np.cross(sigma1, v2))

    # calculate omega (both cases are same)
    omega = math.acos(real_PDB_triangle_correction(np.dot(omega_t1, omega_t2)))
    # determine the sign of omega (+/-)
    sigma1_uni = real_PDB_unit(sigma1)
    sigma1xomega_t1 = np.cross(sigma1, omega_t1)
    sigma1xomega_t2 = np.cross(sigma1, omega_t2)
    omega_dir = real_PDB_unit(np.cross(sigma1xomega_t1, sigma1xomega_t2))
    if np.dot(sigma1_uni, omega_dir) > 0:
        omega = -omega
    else:
        omega = omega

    return theta1, theta2, phi1, phi2, omega, sigma_magnitude







def dtb_PDB_norm_COM(Result: tuple):
    """
    Normalizes the COM of each chain in the given Result and subtracts the interface coordinates of each chain by their respective COM.

    Args:
        Result (9 length tuple): The output result of function(s): 'angle'

    Returns:
        9 length tuple: The tuple contains all the information for further analysis.

    Note:
        The function expects the Result as input which is the output of 'real_PDB_separate_angle()'.
    """

    reaction_chain, new_int_site, new_int_site_distance, unique_chain, COM, angle, normal_point_lst1, normal_point_lst2, one_site_chain = Result
    for i in range(len(unique_chain)):
        for k in range(len(reaction_chain)):
            for j in range(2):
                if unique_chain[i] == reaction_chain[k][j]:
                    for l in range(3):
                        new_int_site[k][j][l] = new_int_site[k][j][l] - COM[i][l]
        for m in range(3):
            COM[i][m] = 0.0
    print('COM is normalized as [0.000, 0.000, 0.000]')
    return reaction_chain, new_int_site, new_int_site_distance, unique_chain, COM, angle, normal_point_lst1, normal_point_lst2, one_site_chain





import numpy as np
import pandas as pd
from .calculate_rmsd import *
from .calculate_gradient import *
from .determine_gagTemplate_structure import *
from .translate_gags_on_sphere import *
from .xyz_to_sphere_coordinates import *
from .restart_pdb_to_df import *
from .gag_write_PDB import *
from ..database_PDB.dtb_PDB_separate_read import *
from ..database_PDB.dtb_PDB_write_PDB import *
from .eliminate_inconsistent_sites import *

def reshape_gag(PathName: str, WritePDB: bool = False):
    """
    This function reshapes the gag lattice recorded experimentally in a PDB file and outputs the standardized sites coordinates of the gag monomers.

    Parameters:
        PathName (str): The full path of the desired PDB file or name of the file if in same directory.
        WritePDB (bool, optional): If True, the function will generate a file in .pdb format containing the standardized COMs 
        and reaction interfaces. The default is False.
    
    Returns:
        finalPositionsVec: The array contains the standardized sites coordinates of the gag monomers.

    """
    # Obtain the coordinates of the COM and interfaces on the gag monomers

    #dtb_PDB_write_PDB(dtb_PDB_separate_read(PathName))
    
    R0 = 65.0           # the target radius of the gag capsid, nm
    distanceCC = 10.0   # the distance between two hexamers, center-to-center distance, nm
    #read gag positions
    #positions = fake_PDB_pdb_to_df("show_structure.pdb")
    positions = fake_PDB_pdb_to_df(PathName)
    positions = positions[["Cite_Name","x_coord", "y_coord", "z_coord"]]

     #convert coordinate unit from angstrom to nm
    positions["x_coord"] = positions["x_coord"]/10.0
    positions["y_coord"] = positions["y_coord"]/10.0
    positions["z_coord"] = positions["z_coord"]/10.0


    # get the number of monomers
    monomer_count = 0
    for i in range(len(positions)):
        if(positions.iloc[i]["Cite_Name"] == "COM"):
            monomer_count += 1
    
    # get count of interfaces for each monomer
    interfaces_count = []
    count = 0
    for i in range(len(positions)):
        count += 1
        if(positions.iloc[i]["Cite_Name"] == "COM"):
            count = 0
        if(i+1 == len(positions) or positions.iloc[i+1]["Cite_Name"] == "COM"):
            interfaces_count.append(count)
    interfaces_count = np.array(interfaces_count)

    # get the index of the COM
    COM_index = []
    curr_index = 0
    for i in range(len(interfaces_count)):
        COM_index.append(curr_index)
        curr_index += (interfaces_count[i]+1)
    


   # eliminate inconsistent sites (sites that are only present on some monomers)
    positions = eliminate_inconsistent_sites(positions, COM_index, interfaces_count)
    consistent_interfaces_count = np.min(interfaces_count)

    # generate a np array to record the coordinates of the COM and interfaces
    positionsVec = np.zeros([monomer_count*(consistent_interfaces_count+1),3])
    for i in range(len(positions)):
        positionsVec[i,:] = [positions.iloc[i]["x_coord"], positions.iloc[i]["y_coord"], positions.iloc[i]["z_coord"]]
    
    
    
    ##############################################
    # find the sphere radius and the sphere center
    # 18 gags, center + 5 nodes' positions, so each gag has 6 positions. I will
    # add the membrane-bind and RNA-bind sites later in this code
    ##############################################

    # first, using the centers of gags to calculate the sphere radius and sphere center 
    numGag = 18
    centersVec = np.zeros([numGag,3])
    for i in range(0,numGag):
        centersVec[i] = positionsVec[6*i]

    sphereXYZR = [0,0,0,70] # initial trial values for sphere x, y, z, and R respectively 
    rmsdOld = calculate_rmsd(centersVec,sphereXYZR)
    isForceSmallEnough = False
    while isForceSmallEnough == False :
        force = calculate_gradient(centersVec,sphereXYZR)
        stepSize = 1.0
        tempXYZR = sphereXYZR - force * stepSize
        rmsdNew = calculate_rmsd(centersVec,tempXYZR)
        while rmsdNew > rmsdOld :
            stepSize = stepSize * 0.8
            tempXYZR = sphereXYZR - force * stepSize
            rmsdNew = calculate_rmsd(centersVec,tempXYZR)
        sphereXYZR = tempXYZR
        rmsdOld = calculate_rmsd(centersVec,sphereXYZR)
        if ( np.linalg.norm(force) < 0.01 ):
            isForceSmallEnough = True

    print('Sphere center position [x,y,z] and radius [R] are, respectively: \n',sphereXYZR) 
    x0 = sphereXYZR[0] 
    y0 = sphereXYZR[1] 
    z0 = sphereXYZR[2] 
    r0 = sphereXYZR[3] 

    ##############################################
    # Second, move the center of the sphere to the axis origin, equivalently move the gags 
    positionsVec[:,0] = positionsVec[:,0] - x0
    positionsVec[:,1] = positionsVec[:,1] - y0
    positionsVec[:,2] = positionsVec[:,2] - z0
    centersVec[:,0] = centersVec[:,0] - x0
    centersVec[:,1] = centersVec[:,1] - y0
    centersVec[:,2] = centersVec[:,2] - z0

    ##############################################
    # Third, move the centers of gag to the sphere surface
    for i in range (0,numGag):
        center = centersVec[i,:]
        move = center/np.linalg.norm(center) * r0 - center
        centersVec[i,:] = centersVec[i,:] + move
        for j in range (0,6):
            positionsVec[j+6*i,:] = positionsVec[j+6*i,:] + move

    ##############################################
    # Fourth, determine the template of the gag (automatically as the first gag). Other gags will be got by translation and rotation of this template gag
    # gagTemplate is the positions of the gag center and five interfaces
    # gagTemplateInterCoeffs is the coefficients of the gag 5 interfaces in the internal basis system
    gagTemplate = determine_gagTemplate_structure(numGag, positionsVec)

    ##############################################
    # Fifth, adjust the hexmerGags 0, 3, 6, 9, 12, 15 
    center0 = centersVec[0,:]   
    center3 = centersVec[3,:]   
    center6 = centersVec[6,:]   
    center9 = centersVec[9,:]
    center12 = centersVec[12,:]
    center15 = centersVec[15,:]
    hexmerCenter = (center0 + center9) / 2.0 

    # the hexamerCenter is almost along the Z axis, just set as along Z axis, then it would be easier for the following rotation and translation of gags
    hexmerCenter = np.zeros([3])
    hexmerCenter[0] = 0.0
    hexmerCenter[1] = 0.0
    hexmerCenter[2] = center0[2]
    # set up the internal coordinate system of the first gag: 3 basis vecs: interBaseVec0, interBaseVec1, interBaseVec2
    interBaseVec0 = center0 / np.linalg.norm(center0)                   # along the radius diraction
    interBaseVec1 = center0 - hexmerCenter
    interBaseVec2 = np.cross(interBaseVec0,interBaseVec1)               # along the tangent line of the hexamer circumference
    interBaseVec2 = interBaseVec2 / np.linalg.norm(interBaseVec2) 
    interBaseVec1 = np.cross(interBaseVec2,interBaseVec0)
    interBaseVec1 = interBaseVec1 / np.linalg.norm(interBaseVec1)
    # calculate gag1's (also gagTemplate) coordinates in its internal coordinate-system
    interCoords = np.zeros([5,3]) # 5 sites, each needs 3 coordinates
    for i in range (0,5) :
        p = gagTemplate[i+1,:] - gagTemplate[0,:] 
        A = np.array([interBaseVec0, interBaseVec1, interBaseVec2])
        interCoords[i,:] = np.dot(p, np.linalg.inv(A))
    sphereCoords0 = xyz_to_sphere_coordinates(center0)
    theta = sphereCoords0[0]
    phi   = sphereCoords0[1]
    r     = sphereCoords0[2]
    r     = R0     # change the radius as the one inputted
    theta = np.arcsin(r0*np.sin(theta)/r) # recaculate the theta angle according to the target radius of the sphere
    # the postions of gags 3, 6, 9, 12, 15 are generated by rotating gag0 along the z-axis
    deltaAngle = 2.0 * np.pi / 6.0
    sixCenters = np.zeros([6,3]) # to store the centers of the hexamer gags
    for i in range (0,6) :
        thetai = theta
        phii = phi + i*deltaAngle
        ri = r
        sixCenters[i,:] = [ri*np.sin(thetai)*np.cos(phii), ri*np.sin(thetai)*np.sin(phii), ri*np.cos(thetai)]

    hexmerCenter[0] = 0  # since the radius is rebuilt by the target radius R0, then the center of the hexamer needs also updated
    hexmerCenter[1] = 0
    hexmerCenter[2] = sixCenters[0,2]
    hexmerPositionsVec = np.zeros([6*6,3]) # to store the postions of the hexamer gags, both their center position and interface position
    for i in range (0,6):
        center = sixCenters[i,:]
        vec1 = center/np.linalg.norm(center) # set up the internal coord system, the three basis vectors. The way is similar as for the template 
        vec2 = center - hexmerCenter
        vec3 = np.cross(vec1,vec2)
        vec3 = vec3/np.linalg.norm(vec3)
        vec2 = np.cross(vec3,vec1)
        vec2 = vec2/np.linalg.norm(vec2)
        interfaces = np.zeros([5,3])
        for j in range (0,5):   # using the template internal coords to determine the five interfaces of each gag
            interfaces[j,:] = interCoords[j,0]*vec1 + interCoords[j,1]*vec2 + interCoords[j,2]*vec3 + center
        hexmerPositionsVec[6*i+0,:] = center
        hexmerPositionsVec[6*i+1:6*i+6,:] = interfaces
        
    # store the hexmerPositionsVec to the new positions
    newPositionsVec = np.zeros([numGag*6,3])

    newPositionsVec[6*0:6*0+6,:]   = hexmerPositionsVec[0:6,:]    # gag0
    newPositionsVec[6*3:6*3+6,:]   = hexmerPositionsVec[6:12,:]   # gag3      
    newPositionsVec[6*6:6*6+6,:]   = hexmerPositionsVec[12:18,:]  # gag6
    newPositionsVec[6*9:6*9+6,:]   = hexmerPositionsVec[18:24,:]  # gag9   
    newPositionsVec[6*12:6*12+6,:] = hexmerPositionsVec[24:30,:]  # gag12
    newPositionsVec[6*15:6*15+6,:] = hexmerPositionsVec[30:36,:]  # gag15

    ##############################################
    # Sixth, find the positions of gag 1,2,4,5,7,8,10,11,13,14,16,17, by moving the hexamer on the sphere surface
    # determine the target position that the original hexamer will move to. I use spherical coordinates
    moveVec = positionsVec[6*2,:] - positionsVec[6*12,:]
    toPosition = moveVec + np.array([0,0,r])
    sphereCrds = xyz_to_sphere_coordinates(toPosition) 
    phi = sphereCrds[1] 
    theta = distanceCC/r # the hexamer center moves 10 nm on the sphere

    fromPosition = np.array([0,0,r]) # center of the original hexamer
    deltaAngle = 2.0 * np.pi / 6.0
    hexmer = hexmerPositionsVec

    # gag 1, 2
    phi = phi
    fromPosition = fromPosition 
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)]) # center of the newhexamer
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*1:6+6*1,:] = newhexmer[6*3:6+6*3,:]     # gag 1 comes from the new position of hexmer gag3 
    newPositionsVec[6*2:6+6*2,:] = newhexmer[6*4:6+6*4,:]     # gag 2 comes from the new position of hexmer gag4

    # gag 4, 5
    phi = phi + deltaAngle
    fromPosition = fromPosition
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)])
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*4:6+6*4,:] = newhexmer[6*4:6+6*4,:]     # gag 4 comes from the new position of hexmer gag4 
    newPositionsVec[6*5:6+6*5,:] = newhexmer[6*5:6+6*5,:]     # gag 5 comes from the new position of hexmer gag5

    # gag 7, 8
    phi = phi + deltaAngle
    fromPosition = fromPosition
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)])
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*7:6+6*7,:] = newhexmer[6*5:6+6*5,:]     # gag 7 comes from the new position of hexmer gag5 
    newPositionsVec[6*8:6+6*8,:] = newhexmer[6*0:6+6*0,:]     # gag 8 comes from the new position of hexmer gag0

    # gag 10, 11
    phi = phi + deltaAngle
    fromPosition = fromPosition
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)])
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*10:6+6*10,:] = newhexmer[6*0:6+6*0,:]     # gag 10 comes from the new position of hexmer gag0 
    newPositionsVec[6*11:6+6*11,:] = newhexmer[6*1:6+6*1,:]     # gag 11 comes from the new position of hexmer gag1

    # gag 13, 14
    phi = phi + deltaAngle
    fromPosition = fromPosition
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)])
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*13:6+6*13,:] = newhexmer[6*1:6+6*1,:]     # gag 13 comes from the new position of hexmer gag1 
    newPositionsVec[6*14:6+6*14,:] = newhexmer[6*2:6+6*2,:]     # gag 14 comes from the new position of hexmer gag2

    # gag 16, 17
    phi = phi + deltaAngle
    fromPosition = fromPosition
    toPosition = np.array([r*np.sin(theta)*np.cos(phi), r*np.sin(theta)*np.sin(phi),r*np.cos(theta)])
    newhexmer = translate_gags_on_sphere(hexmerPositionsVec, fromPosition, toPosition)
    newPositionsVec[6*16:6+6*16,:] = newhexmer[6*2:6+6*2,:]     # gag 16 comes from the new position of hexmer gag2 
    newPositionsVec[6*17:6+6*17,:] = newhexmer[6*3:6+6*3,:]     # gag 17 comes from the new position of hexmer gag3

    ##############################################
    # seventh, add the membrane-bind and RNA-bind sites, then each gag has 8 points
    finalPositionsVec = np.zeros([8*18,3])
    for i in range(0,numGag) :
        center = newPositionsVec[6*i,:]
        surfaceSite = center + 1.0*center/np.linalg.norm(center) # the distance between surfacesite and the center is set as 1nm
        rnaSite = center - 1.0*center/np.linalg.norm(center) # the distance between rnasite and the center is set as 1nm
        finalPositionsVec[0+8*i,:] = center
        finalPositionsVec[1+8*i,:] = surfaceSite
        finalPositionsVec[2+8*i,:] = rnaSite
        finalPositionsVec[3+8*i:8+8*i,:] = newPositionsVec[1+6*i:6+6*i,:]


    # %%
    ###########################################################################################
    ###########################################################################################
    # output the coordinates of each gag
    gagNames = ["A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R"]
    for i in range (0,numGag):
        print(gagNames[i],'\n')
        positions = finalPositionsVec[8*i:8+8*i,:]
        print(positions,'\n')
    
    if WritePDB:
        gag_write_PDB(finalPositionsVec)

    return finalPositionsVec


import math
from ..gen_platonic.distance import distance


def octa_face_leg_reduce(COM: float, leg: float, sigma: float):
    """Reduces the length of an octahedron face leg by a given reduction factor sigma.

    Args:
        COM (float): The center of mass (COM) coordinate of the octahedron face.
        leg (float): The leg vector of the octahedron face.
        sigma (float): The reduction factor for the leg length.

    Returns:
        List[float]: A list of reduced leg coordinates after applying the reduction factor.
            The list contains three floating point values representing the x, y, and z coordinates
            of the reduced leg vector.

    Example:
        COM = [0.0, 0.0, 0.0]
        leg = [1.0, 2.0, 3.0]
        sigma = 0.5
        leg_red = octa_face_leg_reduce(COM, leg, sigma)
        print(leg_red)

    Note:
        The function uses the math module to perform mathematical calculations. The reduction factor
        sigma determines how much the length of the leg vector should be reduced. The leg vector is
        reduced by scaling it with a ratio calculated based on the reduction factor and the distance
        between the center of mass (COM) and the leg vector. The resulting reduced leg coordinates are
        rounded to a given number of decimal places (n) before being returned as a list of floating
        point values.
    """
    n = 12
    angle = math.acos(-1/3)
    red_len = sigma/(2*math.sin(angle/2))
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], n))
    return leg_red






class ProteinComplex():
    """This object contains all the functions (as methods) necessary to edit a protein complex input using a .pdb file downloaded from a database. Then once you are done editing, 
    it can output it as  as NERDSS input, a new .pdb file, or a 3D graph of the complex.
    """


    def __init__(self,FileName: str,ChainsIncluded: list = [None]):
        """Initilizes the ProteinComplex object by reading off of a .pdb file

        Args:
            Filename (str): The full path of the desired PDB file or name of the file if in same directory. 
            ChainsIncluded (list, optional): A list of which chains you want to be included. MUST BE MORE THAN 2!
        """
        from . import dtb_PDB_separate_read

        if len(ChainsIncluded) >= 2 or ChainsIncluded == [None]:
            self.reaction_chain, self.int_site, self.int_site_distance, self.unique_chain, self.COM = dtb_PDB_separate_read(FileName,ChainsIncluded)
            self.one_site_chain = ['na']
        else:
            raise Exception('The ChainsIncluded list, if included, must be greater then 2')


    ## EDITS DATA ##

    def calc_angle(self,NormVector: list = [0.,0.,1.]):
        """This function calculates the 5 associating angles of each pair of interfaces.
        The default normal vector will be assigned as (0, 0, 1). If the co-linear issue occurs, 
        the system will use (0, 1, 0) instead to resolve co-linear issue. The calculated 5 angles 
        will be shown on the screen automatically.
        

        Args:
            NormVector (list, optional): The normal vector used to calculate the angles
            ThrowError(bool, optional): If a co-linear or syntax error occurs, whether 
            it will continue or stop the program


        """
        from . import dtb_PDB_calc_angle

        op = dtb_PDB_calc_angle((self.reaction_chain, self.int_site, self.int_site_distance, 
                                      self.unique_chain, self.COM),NormVector)
        if op[0] == False:
            return False
        else:
            self.set_self_from_tuple(op)
            return True
    


    def norm_COM(self):
        """Normalizes the COM of each chain in the given Result and subtracts the interface coordinates of each chain by their respective COM.
        
        If calc_angles has NOT been run, this function will fail
        """
        from . import dtb_PDB_norm_COM 

        if self.one_site_chain == ["na"]:
            raise Exception("In order to run this function, you must have previously run calc_angle.")

        op = dtb_PDB_norm_COM((self.reaction_chain, self.int_site, self.int_site_distance, 
                                    self.unique_chain, self.COM, self.angle, self.normal_point_lst1, 
                                    self.normal_point_lst2, self.one_site_chain))
        self.set_self_from_tuple(op)



    def filter(self,ChainList):
        """This function will filter the desired chain according to the input list of chain and exclude all the 
            unnecessary coordinate information for future analysis.
        Args:
            ChainList (list): The desired name of chains that users intend to examine. 

        This function must be run before calc_angles or it will fail.
        """
        from . import dtb_PDB_filter

        if self.one_site_chain != ["na"]:
            raise Exception("This function must be run before calc_angles")


        op =  dtb_PDB_filter((self.reaction_chain, self.int_site, self.int_site_distance, self.unique_chain, self.COM),ChainList)

        self.set_self_from_tuple(op)



    def change_sigma(self,ChangeSigma: bool = False, SiteList: list = [], NewSigma: list = []):
        """This function allows users to change the value of sigma (the distance between two binding interfaces). 
        The new sigma value and the corresponding coordinates of interfaces will be shown on the screen and the 
        returns will contain all the information for further analysis. 

        Args:
            ChangeSigma (bool, optional): If True, the users are capable of changing the sigma value; 
                                        if False, the sigma will remain as the original ones. 
            SiteList (list, optional): It consists of the serial numbers of the pair of interfaces for which 
                                    the user needs to modify the sigma value. The serial number is determined 
                                    by the pairing sequence shown by the function ‘real_PDB_separate_read’. 
                                    The serial number should be no greater than the total number of interface 
                                    pairs and no smaller than 0. If the serial number is 0, it means to change 
                                    all pairs of interfaces into a same sigma value.
            NewSigma (list, optional): It consists of the actual sigma value that users desire to change, according 
                                    to the sequence of input ‘SiteList’. 

        This function must be run before calc_angles or it will fail.
        """
        from . import dtb_PDB_change_sigma

        if self.one_site_chain != ["na"]:
            raise Exception("This function must be run before calc_angles")


        op = dtb_PDB_change_sigma((self.reaction_chain, self.int_site, self.int_site_distance, 
                                    self.unique_chain, self.COM),ChangeSigma,SiteList,NewSigma)
        self.set_self_from_tuple(op)
    

    ## OUTPUTS DATA ##

    def write_input(self):
        """Generates a PDB file containing the calculated COMs and reaction interfaces for visualization and comparison with the 
        original PDB file. The input must be the output result of the 'real_PDB_separate_read' function. Note that the unit for 
        the coordinates in the PDB file is Angstrom, not nm, so the values will be 10 times larger than those in NERDSS input 
        files.

        If calc_angles has NOT been run, this function will fail
        """
        from . import dtb_PDB_write_input

        if self.one_site_chain == ["na"]:
            raise Exception("In order to run this function, you must have previously run calc_angle.")
        
        dtb_PDB_write_input((self.reaction_chain, self.int_site, self.int_site_distance, 
                                    self.unique_chain, self.COM, self.angle, self.normal_point_lst1, 
                                    self.normal_point_lst2, self.one_site_chain))



    def plot_3D(self):
        """Generate a 3D plot to display the spatial geometry of each simplified chain.
        """
        from . import dtb_PDB_3D_plot

        dtb_PDB_3D_plot((self.reaction_chain, self.int_site, self.int_site_distance, 
                                    self.unique_chain, self.COM))
        


    def write_PDB(self):
        """Generate a 3D plot to display the spatial geometry of each simplified chain.
        """
        from . import dtb_PDB_write_PDB

        dtb_PDB_write_PDB((self.reaction_chain, self.int_site, self.int_site_distance, 
                                    self.unique_chain, self.COM))


    ## GENERAL METHODS ###

    def set_self_from_tuple(self,op: tuple):
        """Will turn outputted tuple into attributes of this object

        Args:
            op (tuple): The output of one of the functions
        """

        if len(op) == 5:
            self.reaction_chain, self.int_site, self.int_site_distance, self.unique_chain, self.COM = op
        elif len(op) == 9:
            self.reaction_chain, self.int_site, self.int_site_distance, self.unique_chain, self.COM, self.angle, self.normal_point_lst1, self.normal_point_lst2, self.one_site_chain = op
        else:
            raise Exception('The tuple must have a length of 5 or 9')





import math
from ..gen_platonic.mid_pt import mid_pt


def dode_face_COM_coor(a: float, b: float, c: float, d: float, e: float):
    """
    Calculates the center of mass (COM) coordinates for a dodecahedron face
    based on five input coordinates on the same face, and checks for overlap.

    Args:
        a (float): The first coordinate on the face.
        b (float): The second coordinate on the face.
        c (float): The third coordinate on the face.
        d (float): The fourth coordinate on the face.
        e (float): The fifth coordinate on the face.

    Returns:
        list: A list of three float values representing the X, Y, and Z coordinates
        of the center of mass (COM) if the calculated COM coordinates are not overlapped.
        Otherwise, returns the COM coordinates based on the first input coordinate.

    Raises:
        None.

    Example:
        >>> dode_face_COM_coor([0.0, 0.0, 0.0], [1.0, 1.0, 1.0],
        ...                   [2.0, 2.0, 2.0], [3.0, 3.0, 3.0], [4.0, 4.0, 4.0])
        [0.29389262614624, 0.29389262614624, 0.29389262614624]

    Note:
        - The function calculates the center of mass (COM) coordinates by taking
          the midpoint between input coordinates, applying a transformation with
          a scaling factor based on a sine function, and rounding the result to
          14 decimal places.
        - The function checks for overlap among the calculated COM coordinates
          and returns the COM coordinates based on the first input coordinate if
          there is overlap.
    """
    
    # calculate the center of mass(COM) according to 5 coords on the same face
    n = 10
    mid_a = mid_pt(c, d)
    mid_b = mid_pt(d, e)
    mid_c = mid_pt(a, e)
    COM_a = []
    COM_b = []
    COM_c = []
    # calculate 3 COM here and check if they are overlapped
    for i in range(0, 3):
        COM_a.append(round(a[i] + (mid_a[i] - a[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
        COM_b.append(round(b[i] + (mid_b[i] - b[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
        COM_c.append(round(c[i] + (mid_c[i] - c[i]) /
                     (1+math.sin(0.3*math.pi)), 14))
    # checking overlap
    if round(COM_a[0], n) == round(COM_b[0], n) and round(COM_b[0], n) == round(COM_c[0], n) and \
        round(COM_a[1], n) == round(COM_b[1], n) and round(COM_b[1], n) == round(COM_c[1], n) and \
            round(COM_a[2], n) == round(COM_b[2], n) and round(COM_b[2], n) == round(COM_c[2], n):
        return COM_a
    else:
        return COM_a





import numpy as np


def hist_to_csv_old(FullHist: list, SpeciesList: list, OpName: str):
    """WARNING: THIS FUNCTION CURRENTLY HAS TO BE GIVEN FILES OF THE SAME SIZE OR ELSE IT WILL BREAK. (if given multiple)
    Creates a .csv (spreadsheet) file from a histogram.dat file (multi-species)

    Args:
        FullHist (list): holds all of the histogram data
        OpName (str): name of the outputted .csv file

    Returns:
        histogram.csv file: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
        is the number of that complex molecule at that time stamp.
    
    """
    
    column_list = [] #holds the name of each column (Time + each complex name)
    time_list = [] #holds each time. Index corresponds to a sublist in name_count_dict_List
    name_count_dict_list = [] #holds each name/count. Index corresponds with size_list

    #determine longest file
    length = 0
    for file_index,file in enumerate(FullHist):
        if len(file) > length:
            length = len(file)
            length_index = file_index 
    

    #create list with dictionaries for each timestep
    for na in range(0,length+1):
        name_count_dict_list.append({})


    #Creates list with every complex type/size
    for file_index,file in enumerate(FullHist):
        
        for time_index,time in enumerate(file):

            #get timestamps
            if time[0] not in time_list:
                time_list.append(time[0])

            #get complex names and counts
            for complex_index,complexes in enumerate(time[1:]):
                
                #get name
                name = ""
                for index_sp,species in enumerate(complexes[:-1]):
                    name = f"{name}{SpeciesList[index_sp]}: {species}.  "
                    if name not in name_count_dict_list[time_index].keys():
                        name_count_dict_list[time_index][name] = []
                        
                #get count
                name_count_dict_list[time_index][name].append(complexes[-1])

                #creates a list of every 'name'
                if name not in column_list:
                    column_list.append(name)

    #takes mean of each complex type at each timestep
    for time_index,time in enumerate(name_count_dict_list):
        for key in time.keys():
            
            #will add 0s to a list if it does not have data from each file.
            if len(time[key]) < len(FullHist):
                for na in range(len(time[key]), len(FullHist)):
                    time[key].append(0)

            #takes mean of list
            name_count_dict_list[time_index][key] = np.round(np.mean(time[key]),5)
            



    #write the file!
    with open(f'{OpName}.csv', 'w') as write_file:
        
        #create column names
        head = 'Time(s):'
        for column in column_list:
            head += ','
            head += column
        head += '\n'
        write_file.write(head)

        #write the bulk of the file
        for index,timestep in enumerate(time_list):
            
            #initilize writing
            write = ''

            #write time to string
            write += f"{str(timestep)}"

            #write data to string
            for column in column_list:
                write += ','
                if column in name_count_dict_list[index].keys():
                    write += str(name_count_dict_list[index][column])
                else:
                    write += '0'
            
            #commit to file
            write += '\n'
            write_file.write(write)






import numpy as np
import pandas as pd


def translate_gags_on_sphere(hexmer, center1, center2): 
    """move a hexamer gags on the sphere surface
       express the hexmer in the internal coord system, which is based on the point 'from' (center1)

    Args:
        hexmer (_type_): _description_
        center1 (_type_): _description_
        center2 (_type_): _description_

    Returns:
        _type_: _description_
    """

    # set up the internal coordinate basis for hexmer: vec1, vec2, vec3
    vec1 = center1 / np.linalg.norm(center1) # vec1 is along the radius direction
    vec2 = center2 - center1                 # vec2 considers the translational direction
    vec3 = np.cross(vec1,vec2)               # vec3 is along the tangent line of the hexmer
    vec3 = vec3/np.linalg.norm(vec3)
    vec2 = np.cross(vec3,vec1)
    vec2 = vec2/np.linalg.norm(vec2)
    coeffs = np.zeros([hexmer.shape[0],3])
    for i in range (0,hexmer.shape[0]) :
        b = hexmer[i,:] - center1
        A = np.array([vec1, vec2, vec3])
        coeff = [0.0, 0.0, 0.0]
        if (np.linalg.norm(b) > 1e-10) :    
            coeff = np.dot(b, np.linalg.inv(A))
        coeffs[i,:] = coeff
    # move the internal coord system to the point 'to' (center2)
    # find the internal coord system based on the point 'to'
    vec1 = center2 / np.linalg.norm(center2)
    vec2 = center2 - center1 
    vec3 = np.cross(vec1,vec2)
    vec3 = vec3/np.linalg.norm(vec3)
    vec2 = np.cross(vec3,vec1)
    vec2 = vec2/np.linalg.norm(vec2)
    # rebuid the hexmer at this new point 'to'
    newhexmer = np.zeros([hexmer.shape[0],3])
    for i in range (0,hexmer.shape[0]) :
        newhexmer[i,:] = coeffs[i,0]*vec1 + coeffs[i,1]*vec2 + coeffs[i,2]*vec3 + center2
    return newhexmer






import math
from ..gen_platonic.mid_pt import mid_pt


def octa_face_COM_coord(a: float, b: float, c: float):
    """Calculate the center of mass (COM) coordinates for an octahedron face.

    Given the coordinates of three vertices of an octahedron face (a, b, c), this function
    calculates the center of mass (COM) coordinates for that face using the midpoint formula
    and a correction factor based on the sine of 30 degrees.

    Args:
        a (float): The coordinates of the first vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.
        b (float): The coordinates of the second vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.
        c (float): The coordinates of the third vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.

    Returns:
        list: A list of three floats representing the x, y, and z coordinates of the center of mass
              (COM) for the octahedron face.

    Example:
        To calculate the center of mass coordinates for an octahedron face with vertices
        a = [1.0, 2.0, 3.0], b = [4.0, 5.0, 6.0], and c = [7.0, 8.0, 9.0]:
        >>> octa_face_COM_coord([1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0])
        [3.5, 4.5, 5.5]
    """
    mid_a = mid_pt(b, c)
    mid_b = mid_pt(a, c)
    mid_c = mid_pt(a, b)
    COM_a = []
    COM_b = []
    COM_c = []
    for i in range(0, 3):
        COM_a.append(round(a[i] + (mid_a[i] - a[i]) /
                     (1+math.sin(30/180*math.pi)), 12))
        COM_b.append(round(b[i] + (mid_b[i] - b[i]) /
                     (1+math.sin(30/180*math.pi)), 12))
        COM_c.append(round(c[i] + (mid_c[i] - c[i]) /
                     (1+math.sin(30/180*math.pi)), 12))
    if COM_a == COM_b and COM_b == COM_c:
        return COM_a
    else:
        return COM_a





def read_restart(file_name_restart):
    """Read restart file and extract information about complexes and their components.

    Args:
        file_name_restart (str): the name of the restart file to be read.

    Returns:
        complex_lst (list): a list of lists, where each sub-list represents a complex and contains the IDs of its components.

    Examples:
        >>> RESTART_read_restart('restart.txt')
        The total number of complexes is 10
        [[1, 2, 3], [4, 5, 6], [7, 8, 9, 10], [11, 12], [13, 14, 15], [16, 17, 18], [19, 20], [21, 22], [23, 24], [25, 26, 27]]
    """
    with open(file_name_restart, 'r') as file:
        status = False
        count = 0
        complex_lst = []
        for line in file.readlines():
            if line == '#All Complexes and their components \n':
                status = True
            if status:
                if count % 8 == 7:
                    info = line.split()
                    temp_lst = []
                    for i in range(len(info)):
                        if i != 0:
                            temp_lst.append(int(info[i]))
                    complex_lst.append(temp_lst)
                count += 1
            if line == '#Observables \n':
                break
    print('The total number of complexes is', len(complex_lst))
    return complex_lst





import numpy as np

# function definition
def calculate_rmsd(centers, xyzR):  
    """
    Calculates the root mean squared deviation (rmsd) to describe how gags' centers are biased from a perfect sphere surface

    Args:
        centers (18*3 list of floats): the x, y, z coordinates of the 18 gag centers
        xyzR (list of floats): the x, y, z coordinates of the center of the sphere and the radius of the sphere

    Returns:
        s (float): the rmsd value
    """
    x0 = xyzR[0]
    y0 = xyzR[1]
    z0 = xyzR[2]
    r0 = xyzR[3]
    s = 0.0
    for i in range(0,centers.shape[0]):
        xi = centers[i,0]
        yi = centers[i,1]
        zi = centers[i,2]
        ri = np.sqrt( (xi-x0)**2 + (yi-y0)**2 + (zi-z0)**2 )
        s = s + (ri-r0)**2
    return s




def tetr_face_coord(radius: float):
    """Generates vertex coordinates of a tetrahedron given the radius of its circumscribed sphere.

    Args:
        radius (float): The radius of the circumscribed sphere of the tetrahedron.

    Returns:
        list: A list of vertex coordinates for the tetrahedron. The list contains 4 sub-lists,
        each representing the coordinates of one vertex. Each sub-list contains 3 floats representing
        the x, y, and z coordinates of the vertex.

    Example:
        >>> tetr_face_coord(1.0)
        [[0.612372, 0.0, -0.353553], [-0.612372, 0.0, -0.353553],
        [0.0, 0.612372, 0.353553], [0.0, -0.612372, 0.353553]]
    """
    scaler = radius/(3/8)**0.5/2
    v0 = [1, 0, -1/2**0.5]
    v1 = [-1, 0, -1/2**0.5]
    v2 = [0, 1, 1/2**0.5]
    v3 = [0, -1, 1/2**0.5]
    VertCoord = [v0, v1, v2, v3]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





from .tetr_face_leg_reduce_coord_gen import tetr_face_leg_reduce_coord_gen
import numpy as np

def tetr_face_input_coord(radius: float, sigma: float):
    """"Generates input coordinates for a tetrahedral face given the radius of its circumscribed sphere
    and a scaling factor sigma.

    Args:
        radius (float): The radius of the circumscribed sphere of the tetrahedron.
        sigma (float): A scaling factor for reducing the coordinates of the tetrahedral face.

    Returns:
        list: A list of input coordinates for the tetrahedral face. The list contains 5 sub-lists,
        each representing the coordinates of one input vector. Each sub-list contains 3 floats
        representing the x, y, and z coordinates of the vector.

    Example:
        >>> tetr_face_input_coord(1.0, 0.5)
        [[0.0, 0.0, 0.0], [-0.5, 0.0, 0.0], [0.0, -0.5, 0.0], [0.0, 0.0, -0.5], [0.0, 0.0, 0.0]]
    """
    coor = tetr_face_leg_reduce_coord_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = coor_[0] - coor_[0]
    lg1 = coor_[1] - coor_[0]
    lg2 = coor_[2] - coor_[0]
    lg3 = coor_[3] - coor_[0]
    n = -coor_[0]
    return [COM, lg1, lg2, lg3, n]





def cube_vert_coord(radius: float):
    """Calculates the coordinates of the vertices of a cube based on the given radius.

    This function calculates the coordinates of the vertices of a cube based on the given radius, using a scaling factor
    calculated as radius divided by the square root of 3.

    Args:
        radius (float): The radius of the cube.

    Returns:
        list: A list containing eight sub-lists, each containing three floating-point values representing the x, y, and z
        coordinates of a vertex of the cube.

    Example:
        cube_vert_coord(1.0)
        # Calculates the coordinates of the vertices of a cube with a radius of 1.0.
        # Returns a list containing eight sub-lists, each containing three floating-point values representing the x, y, and z
        # coordinates of a vertex of the cube.
    """
        
    scaler = radius/3**0.5
    v0 = [1, 1, 1]
    v1 = [-1, 1, 1]
    v2 = [1, -1, 1]
    v3 = [1, 1, -1]
    v4 = [-1, -1, 1]
    v5 = [1, -1, -1]
    v6 = [-1, 1, -1]
    v7 = [-1, -1, -1]
    VertCoord = [v0, v1, v2, v3, v4, v5, v6, v7]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





from .cube_vert_leg_reduce_coor_gen import cube_vert_leg_reduce_coor_gen
import numpy as np

def cube_vert_input_coord(radius: float, sigma: float):
    """Calculates input coordinates for a cube vertex based on the given radius and sigma.

    This function calculates the input coordinates for a cube vertex based on the given radius and sigma, using the
    `cube_vert_leg_reduce_coor_gen` function to generate the coordinates and then performing various calculations on the
    generated coordinates using NumPy.

    Args:
        radius (float): The radius of the cube.
        sigma (float): The sigma value for the cube vertex.

    Returns:
        tuple: A tuple containing five NumPy arrays, each containing three floating-point values representing the x, y,
        and z coordinates of the input coordinates for the cube vertex. The first array represents the center of mass (COM)
        coordinate, the next three arrays represent the three leg coordinates (lg1, lg2, lg3), and the last array
        represents the normalized vector (n) coordinate.


    Example:
        cube_vert_input_coord(1.0, 0.5)
        # Calculates the input coordinates for a cube vertex with a radius of 1.0 and a sigma value of 0.5.
        # Returns a tuple containing five NumPy arrays, each containing three floating-point values representing the x, y,
        # and z coordinates of the input coordinates for the cube vertex.
    """
        
    coor = cube_vert_leg_reduce_coor_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = np.around(coor_[0] - coor_[0], 8)
    lg1 = np.around(coor_[1] - coor_[0], 8)
    lg2 = np.around(coor_[2] - coor_[0], 8)
    lg3 = np.around(coor_[3] - coor_[0], 8)
    n = np.around(coor_[0]/np.linalg.norm(coor_[0]), 8)
    return COM, lg1, lg2, lg3, n





from .octa_face_vert_coord import octa_face_vert_coord
from .octa_face_COM_coord import octa_face_COM_coord


def octa_face_COM_list_gen(radius: float):
    """Generates a list of center of mass (COM) coordinates for the faces of an octahedron.

    Args:
        radius (float): The radius of the octahedron.

    Returns:
        List: A list of COM coordinates for the faces of the octahedron.

    Example:
        coord = octa_face_vert_coord(radius)
        COM_list = octa_face_COM_list_gen(radius)
        print(COM_list)

    Note:
        The octahedron is assumed to be centered at the origin (0,0,0) and aligned with the
        coordinate axes. The function uses the octa_face_vert_coord() function to generate
        the vertex coordinates of the octahedron, and then calculates the center of mass
        coordinates for the faces using the octa_face_COM_coord() function.
    """
    coord = octa_face_vert_coord(radius)
    COM_list = []
    COM_list.append(octa_face_COM_coord(coord[0], coord[2], coord[4]))
    COM_list.append(octa_face_COM_coord(coord[0], coord[3], coord[4]))
    COM_list.append(octa_face_COM_coord(coord[0], coord[3], coord[5]))
    COM_list.append(octa_face_COM_coord(coord[0], coord[2], coord[5]))
    COM_list.append(octa_face_COM_coord(coord[1], coord[2], coord[4]))
    COM_list.append(octa_face_COM_coord(coord[1], coord[3], coord[4]))
    COM_list.append(octa_face_COM_coord(coord[1], coord[3], coord[5]))
    COM_list.append(octa_face_COM_coord(coord[1], coord[2], coord[5]))
    return COM_list





import numpy as np
from .gen.real_PDB_angles import real_PDB_angles
from .gen.real_PDB_norm_check import real_PDB_norm_check
# This function will calculate five necessary angles: theta_one, theta_two, phi_one, phi_two and omega
# Input variables: four coordinates indicating COM and interaction site of two chains
# First created by Yian Qian
# Modified by Mankun Sang on 04/13/2022
#   1) unit of zero vector and length-one vector
#   2) error messages when v // n
#   3) test scripts
# Modified by Yian Qian & Mankun Sang on 04/16/2022
#   0) correct omega calculation when n // sigma
#   1) generalize the sign determination of phi and omega
#   2) created a function for phi cacluation

def dtb_PDB_calc_angle(Result: tuple, NormVector: list = [0.,0.,1.]):
    """
    This function calculates the 5 associating angles of each pair of interfaces.
    The default normal vector will be assigned as (0, 0, 1). If the co-linear issue occurs, 
    the system will use (0, 1, 0) instead to resolve co-linear issue. The calculated 5 angles 
    will be shown on the screen automatically. If user intends to manually input the normal vector, 
    please refer to function ‘real_PDB_UI’, the separated function does not support manual inputs.
    
    Args:
        Result (5 length tuple): The output result of function(s): 'read','filter','sigma', or first five of function(s): 'COM'
        NormVector (list, optional): The normal vector used to calculate the angles
        ThrowError(bool, optional = True): If a co-linear or syntax error occurs, whether 
        it will continue or stop the program. Recommended to keep as True.
    
    Returns:
        9 length Tuple: All the information for further analysis, including the 5 associating angles of each pair of interfaces.
    """

    reaction_chain, new_int_site, new_int_site_distance, unique_chain, COM = Result
    angle = []
    normal_point_lst1 = []
    normal_point_lst2 = []
    for i in range(len(reaction_chain)):
        chain1 = 0
        chain2 = 0
        for j in range(len(unique_chain)):
            if reaction_chain[i][0] == unique_chain[j]:
                chain1 = j
            if reaction_chain[i][1] == unique_chain[j]:
                chain2 = j
            if reaction_chain[i][0] == unique_chain[chain1] and reaction_chain[i][1] == unique_chain[chain2]:
                break
        
        if NormVector == [0.,0.,1.]:
            normal_point_lst1.append([0., 0., 1.])
            if real_PDB_norm_check(normal_point_lst1[-1], COM[chain1], new_int_site[i][0]) == False:
                pass
            else:
                normal_point_lst1.remove(normal_point_lst1[-1])
                normal_point_lst1.append([0., 1., 0.])

            normal_point_lst2.append([0., 0., 1.])
            if real_PDB_norm_check(normal_point_lst2[-1], COM[chain2], new_int_site[i][1]) == False:
                pass
            else:
                normal_point_lst2.remove(normal_point_lst2[-1])
                normal_point_lst2.append([0., 1., 0.])
        else:
            normal_point_lst1.append(NormVector)
            if real_PDB_norm_check(normal_point_lst1[-1], COM[chain1], new_int_site[i][0]) == False:
                pass
            else:
                raise Exception("Value inputted is invalid because of co-linear or syntax issues")

            normal_point_lst2.append(NormVector)
            if real_PDB_norm_check(normal_point_lst2[-1], COM[chain2], new_int_site[i][1]) == False:
                pass
            else:
                raise Exception("Value inputted is invalid because of co-linear or syntax issues")

        inner_angle = real_PDB_angles(COM[chain1], COM[chain2], new_int_site[i][0], new_int_site[i][1], np.array(
            COM[chain1]) + np.array(normal_point_lst1[-1]), np.array(COM[chain2]) + np.array(normal_point_lst2[-1]))
        angle.append([inner_angle[0], inner_angle[1], inner_angle[2],
                      inner_angle[3], inner_angle[4], inner_angle[5]])
        print("Angles for chain " +
              str(unique_chain[chain1]) + " & " + str(unique_chain[chain2]))
        print("Theta1: %.3f, Theta2: %.3f, Phi1: %.3f, Phi2: %.3f, Omega: %.3f" % (
            inner_angle[0], inner_angle[1], inner_angle[2], inner_angle[3], inner_angle[4]))

    # looking for chains possess only 1 inferface.
    reaction_chain_1d = []
    one_site_chain = []
    for i in reaction_chain:
        for j in i:
            reaction_chain_1d.append(j)
    for i in unique_chain:
        if reaction_chain_1d.count(i) == 1:
            one_site_chain.append(i)
    return reaction_chain, new_int_site, new_int_site_distance, unique_chain, COM, angle, normal_point_lst1, normal_point_lst2, one_site_chain





from .tetr_vert_coord import tetr_vert_coord
from .tetr_vert_COM_leg import tetr_vert_COM_leg

def tetr_vert_COM_leg_gen(radius: float):
    """Generates the center of mass (COM) and midpoints of three edges of a tetrahedron for all possible combinations
    of vertices.
    
    Args:
        radius (float): The radius of the tetrahedron's circumsphere.
        
    Returns:
        list: A list of four COM_leg lists for each vertex combination, where each COM_leg list contains four values,
        [COM, lega, legb, legc], rounded to 10 decimal places.
    """
    coord = tetr_vert_coord(radius)
    COM_leg_list = []
    COM_leg_list.append(tetr_vert_COM_leg(
        coord[0], coord[1], coord[2], coord[3]))
    COM_leg_list.append(tetr_vert_COM_leg(
        coord[1], coord[2], coord[3], coord[0]))
    COM_leg_list.append(tetr_vert_COM_leg(
        coord[2], coord[3], coord[0], coord[1]))
    COM_leg_list.append(tetr_vert_COM_leg(
        coord[3], coord[0], coord[1], coord[2]))
    return COM_leg_list





from .icos_face_write import icos_face_write


def icos_face(radius: float, sigma: float):
    """Write an icosahedron face with given radius and sigma to a file.

    This function writes an icosahedron face with the given radius and sigma values
    to a file using the `icos_face_write()` function from the `icos_face_write` module.

    Args:
        radius (float): The radius of the icosahedron face.
        sigma (float): The sigma value for the icosahedron face.

    Returns:
        parm.inp/icos.mol: input files for NERDSS
    """
    icos_face_write(radius, sigma)
    print('File writing complete!')
    return 0





import numpy as np
import matplotlib.pyplot as plt
import warnings
from .read_transition_matrix import read_transition_matrix
from ..file_managment.save_vars_to_file import save_vars_to_file


def associate_prob_asymmetric(FileName: str, FileNum: int, InitialTime: float, FinalTime: float,
                              SpeciesName: str, DivideSize: int = 2, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """ This function plots a line graph representing the probability of association between complexes of different sizes and other complexes of different sizes.

    Args:
        FileName (str): Path to the ‘.dat’ file, which is usually named as ‘transition_complexes_time.dat’.
        FileNum (int): Number of the total input file. If multiple files are provided, their names should obey the naming rule listed below.
        InitialTime (float): Initial time that users desire to examine and plot on the plot. The acceptable range should not be smaller than the starting time or exceed the ending time of simulation.
        FinalTime (float): Final time that users desire to examine. The acceptable range should not be smaller than the value of InitialTime or exceed the ending time of simulation.
        SpeciesName (str): Name of species that users want to examine, which should also be identical to the name written in the input (.inp and .mol) files.
        DivideSize (int, optional): Value that distinguishes the size of the associate complex, for example, if DivideSize = 2, that means the associate events are classified as ‘associate size < 2’, ‘associate size = 2’ and ‘associate size > 2’. Default value is 2.
        ShowFig (bool, optional): If True, the plot will be shown; if False, the plot will not be shown. No matter the plot is shown or not, the returns will remain the same. Default value is True.
        SaveFig (bool, optional): If True, the plot will be saved as a ‘.png’ file in the current directory; if False, the figure will not be saved. Default value is False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        Line graph. X-axis = size of the complex molecule. Y-axis: Associate probability.

    Notes:
        If multiple input files are given, the output plot will be the average value of all files, and an error bar will also be included.
        Naming rule for input files: If single file is provided, the input file should be named as its original name (‘transition_matrix_time.dat’); if multiple files are provided, the name of input file should also include serial number as ‘transition_matrix_time_X.dat’ where X = 1,2,3,4,5…
    """
    warnings.filterwarnings('ignore')
    
    matrix_list = [] #list of matrix from each file
    
    #create name head/tail
    file_name_head = FileName.split('.')[0]
    file_name_tail = FileName.split('.')[1]
    
    #for each transition matrix file input
    for matrix_file_number in range(1, FileNum+1):
        
        #determining file name (if there are multiple or none)
        if FileNum == 1:
            temp_file_name = FileName
        else:
            temp_file_name = file_name_head + '_' + str(matrix_file_number) + '.' + file_name_tail
        
        #reads file and determines difference
        ti_matrix, tf_matrix = read_transition_matrix(
            temp_file_name, SpeciesName, InitialTime, FinalTime)
        matrix = tf_matrix - ti_matrix
        matrix_list.append(matrix)
    
    #magic
    above = []
    equal = []
    below = []
    for k in range(len(matrix_list)):
        above_temp = np.zeros(len(matrix_list[0][0]))
        equal_temp = np.zeros(len(matrix_list[0][0]))
        below_temp = np.zeros(len(matrix_list[0][0]))
        i = 0
        while i < len(matrix_list[k]):
            j = 0
            while j < len(matrix_list[k][i]):
                if i > j:
                    if i % 2 == 0:
                        if j >= (i-1)/2:
                            if i - j == DivideSize:
                                equal_temp[j] += matrix_list[k][i][j]
                            elif i - j > DivideSize:
                                above_temp[j] += matrix_list[k][i][j]
                            else:
                                below_temp[j] += matrix_list[k][i][j]
                    else:
                        if j >= int(i/2):
                            if (i-1)/2 == j:
                                if i - j == DivideSize:
                                    equal_temp[j] += matrix_list[k][i][j]/2
                                elif i - j > DivideSize:
                                    above_temp[j] += matrix_list[k][i][j]/2
                                else:
                                    below_temp[j] += matrix_list[k][i][j]/2
                            else:
                                if i - j == DivideSize:
                                    equal_temp[j] += matrix_list[k][i][j]
                                elif i - j > DivideSize:
                                    above_temp[j] += matrix_list[k][i][j]
                                else:
                                    below_temp[j] += matrix_list[k][i][j]
                j += 1
            i += 1
        above.append(above_temp)
        equal.append(equal_temp)
        below.append(below_temp)
    
    #more magic
    above_prob = []
    equal_prob = []
    below_prob = []
    for i in range(len(above)):
        above_prob_temp = np.array([])
        equal_prob_temp = np.array([])
        below_prob_temp = np.array([])
        for j in range(len(above[0])):
            sum = above[i][j] + equal[i][j] + below[i][j]
            if sum != 0:
                above_prob_temp = np.append(above_prob_temp, above[i][j]/sum)
                equal_prob_temp = np.append(equal_prob_temp, equal[i][j]/sum)
                below_prob_temp = np.append(below_prob_temp, below[i][j]/sum)
            else:
                above_prob_temp = np.append(above_prob_temp, np.nan)
                equal_prob_temp = np.append(equal_prob_temp, np.nan)
                below_prob_temp = np.append(below_prob_temp, np.nan)
        above_prob.append(above_prob_temp)
        equal_prob.append(equal_prob_temp)
        below_prob.append(below_prob_temp)
    
    #transpose each matrix
    above_prob_rev = []
    for i in range(len(above_prob[0])):
        temp = []
        for j in range(len(above_prob)):
            temp.append(above_prob[j][i])
        above_prob_rev.append(temp)
    equal_prob_rev = []
    for i in range(len(equal_prob[0])):
        temp = []
        for j in range(len(equal_prob)):
            temp.append(equal_prob[j][i])
        equal_prob_rev.append(temp)
    below_prob_rev = []
    for i in range(len(below_prob[0])):
        temp = []
        for j in range(len(below_prob)):
            temp.append(below_prob[j][i])
        below_prob_rev.append(temp)
    
    #calculate means and stds devss
    mean_above = []
    mean_equal = []
    mean_below = []
    std_above = []
    std_equal = []
    std_below = []
   
   #for every column (across all files) determine mean / std
    if FileNum != 1: 
        for i in range(len(above_prob_rev)):
            if np.any(above_prob_rev[i] != above_prob_rev[i-1]) or i == 0:
                mean_above.append(np.nanmean(above_prob_rev[i]))
                std_above.append(np.nanstd(above_prob_rev[i]))
            else:
                mean_above.append(mean_above[i-1])
                std_above.append(std_above[i-1])
            
            if np.any(equal_prob_rev[i] != equal_prob_rev[i-1]) or i == 0:
                mean_equal.append(np.nanmean(equal_prob_rev[i]))
                std_equal.append(np.nanstd(equal_prob_rev[i]))
            else:
                mean_equal.append(mean_equal[i-1])
                std_equal.append(std_equal[i-1])

            if np.any(below_prob_rev[i] != below_prob_rev[i-1]) or i == 0:
                mean_below.append(np.nanmean(below_prob_rev[i]))
                std_below.append(np.nanstd(below_prob_rev[i]))
            else:
                mean_below.append(mean_below[i-1])        
                std_below.append(std_below[i-1])
    else:
        for i in range(len(above_prob_rev)):
            mean_above.append(above_prob_rev[i])
            mean_equal.append(equal_prob_rev[i])
            mean_below.append(below_prob_rev[i])

    n_list = list(range(1, 1 + len(matrix_list[0])))
    
    #output variables
    if SaveVars:
        save_vars_to_file({"cmplx_size":n_list,"mean_associate_probability":[mean_above, mean_equal, mean_below],"std":[std_above, std_equal, std_below]})

    #show figure
    if ShowFig:
        errorbar_color_1 = '#c9e3f6'
        errorbar_color_2 = '#ffe7d2'
        errorbar_color_3 = '#d7f4d7'
        plt.plot(n_list, mean_above, 'C0')
        plt.plot(n_list, mean_equal, 'C1')
        plt.plot(n_list, mean_below, 'C2')
        if FileNum != 1:
            plt.errorbar(n_list, mean_above, yerr=std_above,
                         ecolor=errorbar_color_1, capsize=2)
            plt.errorbar(n_list, mean_equal, yerr=std_equal,
                         ecolor=errorbar_color_2, capsize=2)
            plt.errorbar(n_list, mean_below, yerr=std_below,
                         ecolor=errorbar_color_3, capsize=2)
        plt.legend(['Associate Size > ' + str(DivideSize), 'Associate Size = ' +
                    str(DivideSize), 'Associate Size < ' + str(DivideSize)])
        plt.xlabel('Number of ' + str(SpeciesName) + ' in Single Complex')
        plt.ylabel('Probability')
        plt.xticks(ticks=n_list)
        plt.title('Asymmetric Association Probability')
        if SaveFig:
            plt.savefig('associate_probability_asymmetric.png', dpi=500)
        plt.show()
    return n_list, [mean_above, mean_equal, mean_below], [std_above, std_equal, std_below]





def real_PDB_norm_input(normal_point_lst, chain_name, chain_pair1, chain_pair2):
    """
    Takes user input for normal vector of a chain in a pair of chains and adds it to a list.

    Args:
        normal_point_lst (list): A list to which the normal vector will be appended.
        chain_name (str): The name of the chain for which the normal vector is being input.
        chain_pair1 (str): The name of the first pair for which the normal vector is being input.
        chain_pair2 (str): The name of the other pair for which the normal vector is being input.

    Returns:
        list: A list containing the updated normal vector list after appending the new normal vector input by the user.
    """

    normal_point_1_temp = input('Please input normal vector for ' +
                                chain_name + ' in chain ' + chain_pair1 + " & " + chain_pair2 + ' : ')
    normal_point_1_temp = normal_point_1_temp.strip('[').strip(']').split(',')
    normal_point_1_temp_ = []
    for j in normal_point_1_temp:
        normal_point_1_temp_.append(float(j))
    normal_point_lst.append(normal_point_1_temp_)
    return normal_point_lst





from ..gen_platonic.angle_cal import angle_cal
from .cube_face_leg_reduce_coord_gen import cube_face_leg_reduce_coord_gen
from .cube_face_input_coord import cube_face_input_coord


def cube_face_write(radius: float, sigma: float):
    """Writes input parameters and reaction details to a file for cube face-centered simulation.

    Args:
        radius (float): The radius of the cube face-centered structure.
        sigma (float): The sigma value used for simulation.

    Returns:
        parm.inp/cube.mol files: Inputs for NERDSS

    This function writes the input parameters and reaction details for a cube face-centered
    simulation to a file named 'parm.inp'. The function takes the radius and sigma as input
    arguments, and uses them to calculate the required input parameters and reaction details.
    The file 'parm.inp' contains input parameters such as number of iterations, time steps,
    write frequencies, box boundaries, number of molecules, and reaction details for four
    types of cubes (lg1, lg2, lg3, lg4) based on the given radius and sigma values. The
    function uses helper functions 'cube_face_input_coord', 'cube_face_leg_reduce_coord_gen',
    and 'angle_cal' from other modules to calculate the required input parameters.
    """

    COM, lg1, lg2, lg3, lg4, n = cube_face_input_coord(radius, sigma)
    coord = cube_face_leg_reduce_coord_gen(radius, sigma)
    theta1, theta2, phi1, phi2, omega = angle_cal(
        coord[0][0], coord[0][1], coord[1][0], coord[1][1])

    f = open('parm.inp', 'w')
    f.write(' # Input file (cube face-centered)\n\n')
    f.write('start parameters\n')
    f.write('    nItr = 10000000 #iterations\n')
    f.write('    timeStep = 0.1\n')
    f.write('    timeWrite = 10000\n')
    f.write('    pdbWrite = 10000\n')
    f.write('    trajWrite = 10000\n')
    f.write('    restartWrite = 50000\n')
    f.write('    checkPoint = 1000000\n')
    f.write('    overlapSepLimit = 7.0\n')
    f.write('end parameters\n\n')
    f.write('start boundaries\n')
    f.write('    WaterBox = [500,500,500]\n')
    f.write('end boundaries\n\n')
    f.write('start molecules\n')
    f.write('    cube : 200\n')
    f.write('end molecules\n\n')
    f.write('start reactions\n')
    f.write('    cube(lg1) + cube(lg1) <-> cube(lg1!1).cube(lg1!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg2) + cube(lg2) <-> cube(lg2!1).cube(lg2!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg3) + cube(lg3) <-> cube(lg3!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg4) + cube(lg4) <-> cube(lg4!1).cube(lg4!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg1) + cube(lg2) <-> cube(lg1!1).cube(lg2!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg1) + cube(lg3) <-> cube(lg1!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg1) + cube(lg4) <-> cube(lg1!1).cube(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg2) + cube(lg3) <-> cube(lg2!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg2) + cube(lg4) <-> cube(lg2!1).cube(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg3) + cube(lg4) <-> cube(lg3!1).cube(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('end reactions\n')

    f = open('cube.mol', 'w')
    f.write('##\n')
    f.write('# Cube (face-centered) information file.\n')
    f.write('##\n\n')
    f.write('Name = cube\n')
    f.write('checkOverlap = true\n\n')
    f.write('# translational diffusion constants\n')
    f.write('D = [13.0, 13.0, 13.0]\n\n')
    f.write('# rotational diffusion constants\n')
    f.write('Dr = [0.03, 0.03, 0.03]\n\n')
    f.write('# Coordinates\n')
    f.write('COM   ' + str(round(COM[0], 8)) + '   ' +
            str(round(COM[1], 8)) + '   ' + str(round(COM[2], 8)) + '\n')
    f.write('lg1   ' + str(round(lg1[0], 8)) + '   ' +
            str(round(lg1[1], 8)) + '   ' + str(round(lg1[2], 8)) + '\n')
    f.write('lg2   ' + str(round(lg2[0], 8)) + '   ' +
            str(round(lg2[1], 8)) + '   ' + str(round(lg2[2], 8)) + '\n')
    f.write('lg3   ' + str(round(lg3[0], 8)) + '   ' +
            str(round(lg3[1], 8)) + '   ' + str(round(lg3[2], 8)) + '\n')
    f.write('lg4   ' + str(round(lg4[0], 8)) + '   ' +
            str(round(lg4[1], 8)) + '   ' + str(round(lg4[2], 8)) + '\n')
    f.write('\n')
    f.write('# bonds\n')
    f.write('bonds = 4\n')
    f.write('com lg1\n')
    f.write('com lg2\n')
    f.write('com lg3\n')
    f.write('com lg4\n')
    f.write('\n')


# CUBE VERTEX AS COM




def read_file(FileName: str, SpeciesName: str):
    """Will take in a histogram.dat (single-species) and turn it into a list of lists

    Args:
        FileName (str): Path to the histogram.dat file
        SpeciesName (str): The name of the specific species you want to examine. Should be in the .dat file.

    Returns:
        list of lists: Has many lists, where each sub-list is a new time stamp that includes time at list[i][0]. You can find list 
        of number of each complex type in list[i][1]. List of the number of species count in complex in list[i][2].
    """
    
    #general vars 
    hist = [] # main list that holds each timestamp
    hist_temp = [] # holds all info in 1 timestamp.
    hist_conv = [] # number of species in this complex type at 1 timestamp
    hist_count = [] # num of this complex type at 1 timestamp
    
    #eads through the file
    with open(FileName, 'r') as file:
        for line in file.readlines():
            
            #determining what the line holds
            if line[0:4] == 'Time':
                
                #if this is NOT the first run, add the temp lists to the main list
                if hist_count != [] and hist_conv != []:
                    hist_temp.append(hist_count)
                    hist_temp.append(hist_conv)
                    hist.append(hist_temp)
                
                #reset the temp lists
                hist_count = []
                hist_conv = []
                hist_temp = []

                #set time to start of new temp list
                hist_temp.append(float(line.strip('Time (s): ')))
            
            #if the line holds species information
            else:

                #split the line and determine if it has the right species name
                string = '	' + str(SpeciesName) + ': '
                line = line.strip('. \n').split(string)
                if len(line) != 2:
                    raise Exception('invalid species name')
                
                #adds values to the sub - temp list
                else:
                    hist_count.append(int(line[0]))
                    hist_conv.append(int(line[1]))
    
    #if it is the last run, add it in (needs to be here b/c temps are added at start of new time, not end of previous time)
    hist_temp.append(hist_count)
    hist_temp.append(hist_conv)
    hist.append(hist_temp)
    
    return hist





from .octa_vert_COM_leg_gen import octa_vert_COM_leg_gen
from .octa_vert_leg_reduce import octa_vert_leg_reduce

def octa_vert_leg_reduce_coor_gen(radius: float, sigma: float):
    """Generates a list of center of mass (COM) and reduced leg vectors for an octagonal vertex based on a given radius
    and sigma value.

    This function uses the `octa_vert_COM_leg_gen` function to generate a list of COM and leg vectors for an octagonal
    vertex with the given radius. Then, it applies the `octa_vert_leg_reduce` function to reduce the length of the leg
    vectors based on the given sigma value, and stores the reduced COM and leg vectors in a list.

    Args:
        radius (float): The radius of the octagonal vertex.
        sigma (float): The sigma value used for reducing the length of the leg vectors.

    Returns:
        list: A list of lists, where each sublist contains the reduced COM and leg vectors for an octagonal vertex.
        The structure of the list is as follows:
        [
            [COM1, leg_red1_1, leg_red1_2, leg_red1_3, leg_red1_4],
            [COM2, leg_red2_1, leg_red2_2, leg_red2_3, leg_red2_4],
            ...
        ]

    Example:
        radius = 1.0
        sigma = 0.5
        result = octa_vert_leg_reduce_coor_gen(radius, sigma)
        print(result)
    """
    COM_leg_list = octa_vert_COM_leg_gen(radius)
    COM_leg_red_list = []
    for elements in COM_leg_list:
        temp_list = []
        temp_list.append(elements[0])
        i = 1
        while i <= 4:
            temp_list.append(octa_vert_leg_reduce(
                elements[0], elements[i], sigma))
            i += 1
        COM_leg_red_list.append(temp_list)
    return COM_leg_red_list





from ..gen_platonic.angle_cal import angle_cal
from .icos_face_leg_reduce_coord_gen import icos_face_leg_reduce_coord_gen
from .icos_face_input_coord import icos_face_input_coord


def icos_face_write(radius: float, sigma: float):
    """ Writes input parameters for a simulation of an icosahedron face-centered system.

    Args:
        radius (float): Radius of the icosahedron.
        sigma (float): Sigma parameter for the simulation.

    Returns:
        parm.inp/icos.mol: input files for NERDSS

    This function writes input parameters for a simulation of an icosahedron face-centered system
    to a file named 'parm.inp'. The input parameters include simulation settings such as the number
    of iterations, time step, and output frequency, as well as parameters related to the geometry
    and interaction potentials of the system. The input parameters are derived from the given radius
    and sigma values, which are used to calculate other quantities using helper functions
    `icos_face_input_coord`, `icos_face_leg_reduce_coord_gen`, and `angle_cal`.
    """
    COM, lg1, lg2, lg3, n = icos_face_input_coord(radius, sigma)
    coord = icos_face_leg_reduce_coord_gen(radius, sigma)
    theta1, theta2, phi1, phi2, omega = angle_cal(
        coord[0][0], coord[0][2], coord[11][0], coord[11][3])

    f = open('parm.inp', 'w')
    f.write(' # Input file (icosahedron face-centered)\n\n')
    f.write('start parameters\n')
    f.write('    nItr = 10000000 #iterations\n')
    f.write('    timeStep = 0.1\n')
    f.write('    timeWrite = 10000\n')
    f.write('    pdbWrite = 10000\n')
    f.write('    trajWrite = 10000\n')
    f.write('    restartWrite = 50000\n')
    f.write('    checkPoint = 1000000\n')
    f.write('    overlapSepLimit = 7.0\n')
    f.write('end parameters\n\n')
    f.write('start boundaries\n')
    f.write('    WaterBox = [500,500,500]\n')
    f.write('end boundaries\n\n')
    f.write('start molecules\n')
    f.write('    dode : 200\n')
    f.write('end molecules\n\n')
    f.write('start reactions\n')
    f.write('    icos(lg1) + icos(lg1) <-> icos(lg1!1).icos(lg1!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg2) <-> icos(lg2!1).icos(lg2!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg3) + icos(lg3) <-> icos(lg3!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg2) <-> icos(lg1!1).icos(lg2!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg3) <-> icos(lg1!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg3) <-> icos(lg2!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('end reactions\n')

    f = open('icos.mol', 'w')
    f.write('##\n')
    f.write('# Icosahehedron (face-centered) information file.\n')
    f.write('##\n\n')
    f.write('Name = icos\n')
    f.write('checkOverlap = true\n\n')
    f.write('# translational diffusion constants\n')
    f.write('D = [13.0, 13.0, 13.0]\n\n')
    f.write('# rotational diffusion constants\n')
    f.write('Dr = [0.03, 0.03, 0.03]\n\n')
    f.write('# Coordinates\n')
    f.write('COM   ' + str(round(COM[0], 8)) + '   ' +
            str(round(COM[1], 8)) + '   ' + str(round(COM[2], 8)) + '\n')
    f.write('lg1   ' + str(round(lg1[0], 8)) + '   ' +
            str(round(lg1[1], 8)) + '   ' + str(round(lg1[2], 8)) + '\n')
    f.write('lg2   ' + str(round(lg2[0], 8)) + '   ' +
            str(round(lg2[1], 8)) + '   ' + str(round(lg2[2], 8)) + '\n')
    f.write('lg3   ' + str(round(lg3[0], 8)) + '   ' +
            str(round(lg3[1], 8)) + '   ' + str(round(lg3[2], 8)) + '\n')
    f.write('\n')
    f.write('# bonds\n')
    f.write('bonds = 3\n')
    f.write('com lg1\n')
    f.write('com lg2\n')
    f.write('com lg3\n')
    f.write('\n')


# ICOSAHEDRON VERTEX AS COM




from .dode_face_write import dode_face_write


def dode_face(radius: float, sigma: float):
    """
    Generates a dodecahedron face using the given radius and sigma values,
    writes it to a file using `dode_face_write` function, and prints a
    completion message.

    Args:
        radius (float): The radius of the dodecahedron.
        sigma (float): The sigma value to use for generating the dodecahedron.

    Returns:
        parm.inp/cube.mol file: inputs for NERDSS
    """
    
    dode_face_write(radius, sigma)
    print('File writing complete!')
    return 0





import math


def vert_coord(radius: float):
    """Generates the vertex coordinates of an icosahedron face.

    Args:
        radius (float): Radius of the icosahedron.

    Returns:
        list: A list of vertex coordinates of the icosahedron face.

    Example:
        >>> icos_face_vert_coord(1.0)
        [[v0_x, v0_y, v0_z],
         [v1_x, v1_y, v1_z],
         ...
        ]
    """
    scaler = radius/(2*math.sin(2*math.pi/5))
    m = (1+5**0.5)/2
    v0 = [0, 1, m]
    v1 = [0, 1, -m]
    v2 = [0, -1, m]
    v3 = [0, -1, -m]
    v4 = [1, m, 0]
    v5 = [1, -m, 0]
    v6 = [-1, m, 0]
    v7 = [-1, -m, 0]
    v8 = [m, 0, 1]
    v9 = [m, 0, -1]
    v10 = [-m, 0, 1]
    v11 = [-m, 0, -1]
    VertCoord = [v0, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





import numpy as np
import matplotlib.pyplot as plt
import warnings
from .read_transition_matrix import read_transition_matrix
from ..file_managment.save_vars_to_file import save_vars_to_file

def dissociate_prob_symmetric(FileName: str, FileNum: int, InitialTime: float, FinalTime: float,
                              SpeciesName: str, DivideSize: int = 2, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """ Calculates and displays the probability of dissociation of complexes of different sizes into other complexes of different sizes.

    Args:
        FileName (str): Path to the '.dat' file representing the histogram data to be analyzed.
        FileNum (int): Number of input files. If multiple files are provided, their names should obey the naming rule described below.
        InitialTime (float): The initial time that users desire to examine (in seconds).
        FinalTime (float): The final time that users desire to examine (in seconds).
        SpeciesName (str): The name of the species that users want to examine, identical to the name written in the input (.inp and .mol) files.
        DivideSize (int, optional): The value that distinguishes the size of the dissociate complex (default=2). For example, if DivideSize=2, that means the dissociate events are classified as 'dissociate size < 2', 'dissociate size = 2' and 'dissociate size > 2'.
        ShowFig (bool, optional): If True, the plot will be displayed (default=True).
        SaveFig (bool, optional): If True, the plot will be saved as a '.png' file in the current directory (default=False).
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.
        
    Returns:
        A tuple containing the following: 
         - The x-axis data
         - The y-axis data
         - Errors in y-axis data
         - The bin size.
                
    Notes:
        - If a single file is provided, the input file should be named as its original name ('transition_matrix_time.dat'). If multiple files are provided, the name of the input file should include a serial number as 'transition_matrix_time_X.dat' where X = 1, 2, 3, 4, 5, etc.
        - The function generates a line plot of the probability of dissociation of complexes of different sizes into other complexes of different sizes. The x-axis represents the size of the complex, and the y-axis represents the dissociate probability. Three lines will exist in the line graph, representing dissociating to complexes of sizes less than, equal to, or greater than the specified size, respectively. 'Symmetric' in the function name means that for the dissociate reaction, both sizes of complexes are counted as dissociating events symmetrically, for example, if a dissociate event occurs where a heptamer dissociates into a tetramer and a trimer, then this event is counted twice, which are heptamer dissociates to tetramer and heptamer dissociates to trimer. If multiple input files are given, the output plot will be the average value of all files and an error bar will also be included.
    """
    warnings.filterwarnings('ignore')
    
    matrix_list = []
    
    #create name head/tail
    file_name_head = FileName.split('.')[0]
    file_name_tail = FileName.split('.')[1]
    
    #for each transition matrix file input
    for matrix_file_number in range(1, FileNum+1):
        
        #determining file name (if there are multiple or none)
        if FileNum == 1:
            temp_file_name = FileName
        else:
            temp_file_name = file_name_head + '_' + str(matrix_file_number) + '.' + file_name_tail
        
        #reads file and determines difference
        ti_matrix, tf_matrix = read_transition_matrix(
            temp_file_name, SpeciesName, InitialTime, FinalTime)
        matrix = tf_matrix - ti_matrix
        matrix_list.append(matrix)
    
    #magic
    above = []
    equal = []
    below = []
    for k in range(len(matrix_list)):
        above_temp = np.zeros(len(matrix_list[0][0]))
        equal_temp = np.zeros(len(matrix_list[0][0]))
        below_temp = np.zeros(len(matrix_list[0][0]))
        i = 0
        while i < len(matrix_list[k][0]):
            j = 0
            while j < len(matrix_list[k][i]):
                if i > j:
                    if j + 1 == DivideSize:
                        equal_temp[i] += matrix_list[k][j][i]
                    elif j + 1 > DivideSize:
                        above_temp[i] += matrix_list[k][j][i]
                    else:
                        below_temp[i] += matrix_list[k][j][i]
                j += 1
            i += 1
        above.append(above_temp)
        equal.append(equal_temp)
        below.append(below_temp)
    
    #magic
    above_prob = []
    equal_prob = []
    below_prob = []
    for i in range(len(above)):
        above_prob_temp = np.array([])
        equal_prob_temp = np.array([])
        below_prob_temp = np.array([])
        for j in range(len(above[0])):
            sum = above[i][j] + equal[i][j] + below[i][j]
            if sum != 0:
                above_prob_temp = np.append(above_prob_temp, above[i][j]/sum)
                equal_prob_temp = np.append(equal_prob_temp, equal[i][j]/sum)
                below_prob_temp = np.append(below_prob_temp, below[i][j]/sum)
            else:
                above_prob_temp = np.append(above_prob_temp, np.nan)
                equal_prob_temp = np.append(equal_prob_temp, np.nan)
                below_prob_temp = np.append(below_prob_temp, np.nan)
        above_prob.append(above_prob_temp)
        equal_prob.append(equal_prob_temp)
        below_prob.append(below_prob_temp)
    
    #transpose
    above_prob_rev = []
    for i in range(len(above_prob[0])):
        temp = []
        for j in range(len(above_prob)):
            temp.append(above_prob[j][i])
        above_prob_rev.append(temp)
    equal_prob_rev = []
    for i in range(len(equal_prob[0])):
        temp = []
        for j in range(len(equal_prob)):
            temp.append(equal_prob[j][i])
        equal_prob_rev.append(temp)
    below_prob_rev = []
    for i in range(len(below_prob[0])):
        temp = []
        for j in range(len(below_prob)):
            temp.append(below_prob[j][i])
        below_prob_rev.append(temp)
    
    #calculate means and stds devss
    mean_above = []
    mean_equal = []
    mean_below = []
    std_above = []
    std_equal = []
    std_below = []
   
   #for every column (across all files) determine mean / std
    if FileNum != 1: 
        for i in range(len(above_prob_rev)):
            if np.any(above_prob_rev[i] != above_prob_rev[i-1]) or i == 0:
                mean_above.append(np.nanmean(above_prob_rev[i]))
                std_above.append(np.nanstd(above_prob_rev[i]))
            else:
                mean_above.append(mean_above[i-1])
                std_above.append(std_above[i-1])
            
            if np.any(equal_prob_rev[i] != equal_prob_rev[i-1]) or i == 0:
                mean_equal.append(np.nanmean(equal_prob_rev[i]))
                std_equal.append(np.nanstd(equal_prob_rev[i]))
            else:
                mean_equal.append(mean_equal[i-1])
                std_equal.append(std_equal[i-1])

            if np.any(below_prob_rev[i] != below_prob_rev[i-1]) or i == 0:
                mean_below.append(np.nanmean(below_prob_rev[i]))
                std_below.append(np.nanstd(below_prob_rev[i]))
            else:
                mean_below.append(mean_below[i-1])        
                std_below.append(std_below[i-1])
    else:
        for i in range(len(above_prob_rev)):
            mean_above.append(above_prob_rev[i])
            mean_equal.append(equal_prob_rev[i])
            mean_below.append(below_prob_rev[i])

    mean_above = np.nan_to_num(mean_above)
    mean_equal = np.nan_to_num(mean_equal)
    mean_below = np.nan_to_num(mean_below)
    
    n_list = list(range(1, 1 + len(matrix_list[0])))
    
    #output variables
    if SaveVars:
        save_vars_to_file({"cmplx_size":n_list,"mean_dissociate_probability":[mean_above, mean_equal, mean_below],"std":[std_above, std_equal, std_below]})

    #show figure
    if ShowFig:
        errorbar_color_1 = '#c9e3f6'
        errorbar_color_2 = '#ffe7d2'
        errorbar_color_3 = '#d7f4d7'
        plt.plot(n_list, mean_above, 'C0')
        plt.plot(n_list, mean_equal, 'C1')
        plt.plot(n_list, mean_below, 'C2')
        if FileNum != 1:
            plt.errorbar(n_list, mean_above, yerr=std_above,
                         ecolor=errorbar_color_1, capsize=2)
            plt.errorbar(n_list, mean_equal, yerr=std_equal,
                         ecolor=errorbar_color_2, capsize=2)
            plt.errorbar(n_list, mean_below, yerr=std_below,
                         ecolor=errorbar_color_3, capsize=2)
        plt.legend(['Dissociate Size > ' + str(DivideSize), 'Dissociate Size = ' +
                    str(DivideSize), 'Dissociate Size < ' + str(DivideSize)])
        plt.xlabel('Number of ' + str(SpeciesName) + ' in Single Complex')
        plt.ylabel('Probability')
        plt.xticks(ticks=n_list)
        plt.title('Symmetric Dissociation Probability')
        if SaveFig:
            plt.savefig('dissociate_probability_symmetric.png', dpi=500)
        plt.show()
    return n_list, [mean_above, mean_equal, mean_below], [std_above, std_equal, std_below]





import unittest
import numpy as np
from typing import List

def distance(a: List[float], b: List[float]) -> float:
    """Compute the Euclidean distance between two points in n-dimensional space.

    Args:
        a (List[float]): The coordinates of the first point.
        b (List[float]): The coordinates of the second point.

    Returns:
        float: The Euclidean distance between the two points.


    Examples:
        >>> a = [0, 0, 0]
        >>> b = [1, 1, 1]
        >>> distance(a, b)
        1.7320508075688772

    Notes
        This function computes the Euclidean distance between two points by taking
        the square root of the sum of squared differences of each coordinate. The 
        result is rounded to 15 decimal places using string formatting.
    """
    return float(f"{np.linalg.norm(np.array(a) - np.array(b)):.15f}")

class TestDistance(unittest.TestCase):
    def test_distance(self):
        a = [0, 0, 0]
        b = [1, 1, 1]
        self.assertAlmostEqual(distance(a, b), 1.7320508075688772)

        a = [3, 4, 0]
        b = [0, 0, 12]
        self.assertAlmostEqual(distance(a, b), 13.0)

        a = [0, 4]
        b = [3, 0]
        self.assertAlmostEqual(distance(a, b), 5.0)

if __name__ == '__main__':
    unittest.main()



def find_complexes(sorted_complexes, NumDict):
    """Finds the complexes with the correct number of proteins of each type.
    
    Args:
        sorted_complexes (array): holds each complex as a subdictionary. Then each dictionaries key = name of a protein, value = list of protein indexes in that complex
        NumDict (dictionary): A dictionary that holds the requested number of protein types in a complex
    
    Returns:
        A list of integers representing the protein numbers that form the correct size /composition of a complex.
    """


    protein_remain = [] #list of proteins in the complexes (what is returned)
    protein_complex_hits = 0 #how many protein complexes have the correct number
    
    #each row of the dataframe
    for complex in sorted_complexes:

        #if row/protein complex has the correct number of each protein type
        correct_complex_size = True
        for protein_type,protein_list in complex.items():
            if len(protein_list) != NumDict[protein_type]:
                correct_complex_size = False
                break
        
        if correct_complex_size: 
            protein_complex_hits += 1
            
            #add every protein num that is in that complex
            for protein_num in protein_list:
                protein_remain.append(int(protein_num))
    
    #print success / failure
    if protein_complex_hits == 0:
        print('No complexes where found with the specific number of protein types.')
    else:
        print(f'Found {protein_complex_hits} protein complexes!')

    return protein_remain





def real_PDB_data_check(data):
    """
    Check if the PDB data is valid.

    Args:
        data (list): List of PDB data to be checked.

    Returns:
        int: -1 if the length of data is not equal to 12 because the amino acid name combined with previous section
             -2 if the length of data is equal to 12 but the amino acid name is not of length 3.
              1 if the data is valid.

    """
    
    if len(data) != 12:
        if len(data[2]) > 4:
            return -1  # Amino acid name stick with info before
    else:
        if len(data[3]) == 3:
            return 1  # True data
        else:
            return -2  # Wrong amino acid name



from .octa_face_leg_reduce_coord_gen import octa_face_leg_reduce_coord_gen
import numpy as np

def octa_face_input_coord(radius: float, sigma: float):
    """Generates input coordinates for an octahedron face reduction algorithm.

    Args:
        radius (float): The radius of the octahedron.
        sigma (float): The sigma value for the reduction algorithm.

    Returns:
        List: A list of input coordinates for the octahedron face reduction algorithm.
            The list contains the following:
            - COM (numpy.array): The center of mass (COM) coordinates for the octahedron face.
            - lg1 (numpy.array): The first leg vector from COM to vertex 1.
            - lg2 (numpy.array): The second leg vector from COM to vertex 2.
            - lg3 (numpy.array): The third leg vector from COM to vertex 3.
            - n (numpy.array): The normal vector of the octahedron face.

    Example:
        radius = 1.0
        sigma = 0.5
        input_coord = octa_face_input_coord(radius, sigma)
        print(input_coord)

    Note:
        The octahedron is assumed to be centered at the origin (0,0,0) and aligned with the
        coordinate axes. The function uses the octa_face_leg_reduce_coord_gen() function to generate
        the input coordinates for the face reduction algorithm. The input coordinates include the
        center of mass (COM) coordinates, leg vectors, and normal vector of the octahedron face.
    """

    coor = octa_face_leg_reduce_coord_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = coor_[0] - coor_[0]
    lg1 = coor_[1] - coor_[0]
    lg2 = coor_[2] - coor_[0]
    lg3 = coor_[3] - coor_[0]
    n = -coor_[0]
    return [COM, lg1, lg2, lg3, n]





from .cube_face_write import cube_face_write


def cube_face(radius: float, sigma: float):
    """Generates a cube face file for visualization of a molecular system.

    This function generates a cube face file using the provided radius and sigma values, which can be used for
    visualization of a molecular system in a molecular visualization software. The cube face file is written using the
    `cube_face_write` function from the `.cube_face_write` module.

    Args:
        radius (float): The radius of the cube face.
        sigma (float): The sigma value for the cube face.

    Returns:
        parm.inp/cube.mol files: Inputs for NERDSS
    """
    
    cube_face_write(radius, sigma)
    print('File writing complete!')
    return 0





def list_index_exists(lst,index):
    """Checks if a there is something at index of the list

    Args:
        lst (list): the list you are checking
        index (int): the index of the list you are checking

    Returns:
        boolean: if that index exists for that list
    """
    
    try:
        lst[index]
        return True
    except IndexError:
        return False



import numpy as np


def gag_write_PDB(sites_coord: list):
    
    
    read_file = open('show_structure.pdb', 'r')
    write_file = open("regularized_structure.pdb", "w")

    lines = read_file.readlines()
    sites_coord_index = 0
    interface_serial = 6
    for i in range(len(lines)):
        if(lines[i][0:4] != "ATOM"):
            write_file.write(lines[i])
        else:
            if(lines[i][13:17].strip() == "COM"):
                if(interface_serial < 6):
                    for j in range(6- interface_serial):
                        
                        write_line = "extra line\n"
                        write_file.write(str(sites_coord_index) + write_line)
                        sites_coord_index +=1 
                    
                    write_line = lines[i]
                    write_file.write(str(sites_coord_index) + write_line)
                    sites_coord_index += 1
                    interface_serial = 1
                else:
                    
                    write_line = lines[i]
                    write_file.write(str(sites_coord_index) + write_line)
                    sites_coord_index += 1
                    interface_serial = 1
            else:
                
                write_line = lines[i][:7] + format() + sites_coord[sites_coord_index] + lines[i][:54]
                write_file.write(str(sites_coord_index) + write_line)
                sites_coord_index += 1
                interface_serial += 1
        

            

    read_file.close
    write_file.close


    return 0


import numpy as np
from ..gen_platonic.COM_leg_list_gen import COM_leg_list_gen
from .cube_face_COM_leg_list_gen import cube_face_COM_leg_list_gen
from .cube_face_leg_reduce import cube_face_leg_reduce

def cube_face_leg_reduce_coord_gen(radius: float, sigma: float):
    """Generates a list of reduced center of mass and leg vectors for cube faces.

    This function takes the radius and sigma value as inputs, and generates a list of reduced center of mass (COM) and leg
    vectors for the cube faces of a platonic solid. The reduction is performed using the 'cube_face_leg_reduce' function
    from the 'cube_face_leg_reduce' module, and the original COM and leg vectors are obtained from the 'cube_face_COM_leg_list_gen'
    and 'cube_face_COM_leg_list_gen' functions respectively.

    Args:
        radius (float): The radius of the platonic solid.
        sigma (float): The sigma value for the reduction.

    Returns:
        List: Contains reduced COM and leg vectors for the cube faces. Each element in the list is a sublist containing the reduced
        COM vector followed by the reduced leg vectors for each face. The coordinates in the vectors are rounded to 8 decimal places.

    Raises:
        None.

    Example:
        >>> cube_face_leg_reduce_coord_gen(1.0, 0.1)
        [[0.0, [0.131826, 0.131826, 0.131826], [0.131826, 0.131826, -0.131826], [-0.131826, 0.131826, -0.131826], [-0.131826, 0.131826, 0.131826]],
         [0.0, [-0.131826, 0.131826, 0.131826], [-0.131826, -0.131826, 0.131826], [-0.131826, -0.131826, -0.131826], [-0.131826, 0.131826, -0.131826]],
         ...
        ]

    """
    COM_leg_list = cube_face_COM_leg_list_gen(radius)
    COM_leg_red_list = []
    for elements in COM_leg_list:
        temp_list = []
        temp_list.append(np.around(elements[0], 8))
        i = 1
        while i <= 4:
            temp_list.append(np.around(cube_face_leg_reduce(
                elements[0], elements[i], sigma), 8))
            i += 1
        COM_leg_red_list.append(temp_list)
    return COM_leg_red_list





import os
import numpy as np

def save_variable_to_file(variable, file_name=None, file_path=None):
    """Save a list or NumPy array of numbers to a .dat file, one number per line.

    This function takes a list or NumPy array of numbers as input, saves it to a .dat file
    with each number on a new line, and allows for customization of the file name and file
    path. If no file name is provided, the default name is "my_var". If no file path is
    provided, the default path is the current folder.

    Parameters
    ----------
    variable : list or numpy.array
        The list or NumPy array of numbers to be saved.
    file_name : str, optional
        The name of the file, by default "my_var".
    file_path : str, optional
        The abs path of the file, by default the current folder.

    Returns
    -------
    None
    """
    if not isinstance(variable, (list, np.ndarray)):
        raise ValueError("The input variable must be a list or a NumPy array of numbers.")
    
    if isinstance(variable, np.ndarray):
        variable = variable.tolist()
    
    if not all(isinstance(item, (int, float)) for item in variable):
        raise ValueError("All elements in the list or array must be numbers (integers or floats).")
    
    if file_name is None:
        file_name = "my_var" + '.dat'
    else:
        file_name = file_name + '.dat'

    if file_path is None:
        file_path = os.getcwd()
    else:
        file_path = os.path.abspath(file_path)

    file_full_path = os.path.join(file_path, file_name)

    with open(file_full_path, 'w') as f:
        for number in variable:
            f.write(f'{number}\n')



def xyz_to_csv(FileName: str, LitNum: int = -1, OpName: str = "output_file"):
    """Converts a .xyz file to a .csv file for a specific or entire time frame.

    Args:
        FileName (str): The path to the input .xyz file, usually named 'trajectory.xyz'.
        LitNum (int, optional = -1): The number of iterations to examine. If -1, the entire iteration will be extracted.
        OpName (str, optional = “output_file”): The name of the outputted file. 

    Returns:
        A .csv file.

    Description:
        This function enables users to convert the output .xyz file by NERDSS simulation into a .csv file of a specific or entire time frame. The generated csv file will contain 5 columns, including number of iteration, species name, x, y, and z coordinates.

    Sample:
        xyz_to_csv('/Users/UserName/Documents/trajectory.xyz', 100000000) # Extracts iteration 100000000
        xyz_to_csv('/Users/UserName/Documents/trajectory.xyz', -1) # Extracts the entire iteration
    """
    
    #determines which iterations will be included
    write_file_name = f"{OpName}.csv"

    if LitNum != -1:
        lit_switch = False
    else:
        lit_switch = True
    
    #open read and write file
    with open(FileName, 'r') as read_file, open(write_file_name, 'w') as write_file:
        
        #creates header
        head = 'literation,name,x,y,z\n'
        write_file.write(head)
        
        #reads each line
        for line in read_file.readlines():
            
            #determines whether this iteration will be read or not
            if LitNum != -1:
                if line[0:11] == 'iteration: ':
                    if int(line.split(' ')[1]) == LitNum:
                        lit_switch = True
                    else:
                        lit_switch = False
                    literation = LitNum
            else:
                if line[0:11] == 'iteration: ':
                    literation = int(line.split(' ')[1])

            #if reading is enabled
            if lit_switch:
                
                #if it has the correct length
                info = line.strip(' ').strip('\n').split()
                if len(info) == 4:
                    
                    #write the row name
                    write_info = str(literation) + ','
                    
                    for word in info:
                        write_info += word

                        #if it is not last add ',' else \n
                        if word != info[-1]:
                            write_info += ','
                        else:
                            write_info += '\n'
                    write_file.write(write_info)
    return 0





from .dode_vert_coord import dode_vert_coord
from .dode_vert_COM_leg import dode_vert_COM_leg


def dode_vert_COM_leg_gen(radius: float):
    """Generates and returns a list of dodecahedron leg vertices based on the center of mass (COM) and radius.

    Args:
        radius (float): Radius of the dodecahedron.

    Returns:
        list: List of vertices as [COM_leg1, COM_leg2, ..., COM_leg20], where each COM_leg is a list of vertices as [COM, lega, legb, legc], rounded to 10 decimal places.

    Example:
        >>> dode_vert_COM_leg_gen(1.0)
        [
            [COM1, lega1, legb1, legc1],
            [COM2, lega2, legb2, legc2],
            ...
            [COM20, lega20, legb20, legc20]
        ]
    """
     
    coord = dode_vert_coord(radius)
    COM_leg_list = []
    COM_leg_list.append(dode_vert_COM_leg(
        coord[0], coord[1], coord[12], coord[16]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[1], coord[0], coord[13], coord[17]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[2], coord[3], coord[14], coord[18]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[3], coord[2], coord[15], coord[19]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[4], coord[6], coord[12], coord[14]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[5], coord[7], coord[13], coord[15]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[6], coord[4], coord[16], coord[18]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[7], coord[5], coord[17], coord[19]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[8], coord[9], coord[12], coord[13]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[9], coord[8], coord[14], coord[15]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[10], coord[11], coord[16], coord[17]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[11], coord[10], coord[18], coord[19]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[12], coord[0], coord[4], coord[8]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[13], coord[1], coord[5], coord[8]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[14], coord[2], coord[4], coord[9]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[15], coord[3], coord[5], coord[9]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[16], coord[0], coord[6], coord[10]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[17], coord[1], coord[7], coord[10]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[18], coord[2], coord[6], coord[11]))
    COM_leg_list.append(dode_vert_COM_leg(
        coord[19], coord[3], coord[7], coord[11]))
    return COM_leg_list





import numpy as np
import matplotlib.pyplot as plt
from .read_multi_hist import read_multi_hist
from ...file_managment.save_vars_to_file import save_vars_to_file

def multi_heatmap_complex_dist(FileName: str, FileNum: int, InitialTime: float, FinalTime: float, xAxis: str, yAxis: str,
                  SpeciesList: list = [], xBarSize: int = 1, yBarSize: int = 1,
                  ShowFig: bool = True, ShowMean: bool = False, ShowStd: bool = False, SaveFig: bool = False, SaveVars: bool = False):
    """ Creates a 3D heatmap from a histogram.dat (multi-species) that shows distrubution of sizes of selected species.

    Args:
        FileName (str): Path to the histogram.dat file
        FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
        InitialTime (float): The starting time. Must not be smaller / larger then times in file.
        FinalTime (float): The ending time. Must not be smaller / larger then times in file.
        xAxis (str): Species shown on X-axis.
        yAxis (str): Species shown on Y-axis.
        SpeciesList (list, optional): The names of the species you want to examine. Should be in the .dat file.
        xBarSize (int, optional): The size of each data bar in the X-dimension. Defaults to 1.
        yBarSize (int, optional): The size of each data bar in the Y-dimension. Defaults to 1.
        ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
        ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
        ShowFig (bool, optional): If the plot is shown. Defaults to True.
        SaveFig (bool, optional): If the plot is saved. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        3D Histogram. X-axis / Y-axis: the distribution of sizes of each specified species. Color: relative occurance of each complex.
    """
    file_name_head = FileName.split('.')[0]
    file_name_tail = FileName.split('.')[1]
    count_list_sum = []

    #create species list 
    SpeciesList = []
    SpeciesList.append(xAxis)
    SpeciesList.append(yAxis)

    #true max X and Y (across all files)
    true_max_x = 0
    true_max_y = 0

    #runs through every file
    for histogram_file_number in range(1, FileNum+1):
        
        #determining file name (if there are multiple or none)
        if FileNum == 1:
            temp_file_name = FileName
        else:
            temp_file_name = file_name_head + '_' + str(histogram_file_number) + '.' + file_name_tail

        max_x_size = 0 #stores the max number of the x protein type in 1 complex (1 file vv)
        max_y_size = 0 #stores the max number of the y protein type in 1 complex
        count_list = [[0]]
        data_count = 0

        #get lists of data from file
        hist_list = read_multi_hist(temp_file_name, SpeciesList=SpeciesList)

        #run through every timestep in the file, to find the max # of Xs and Ys in a single complex
        for time_step in hist_list:
            if time_step != []:
                time = time_step[0]
                if InitialTime <= time <= FinalTime:
                    data_count += 1

                    #run through each protein complex in timestep
                    for protein_complex in time_step[1:]:

                        #find number of each protein type in this complex. Then how many complexes of this type there are
                        x_size = protein_complex[0]
                        x_size = int(x_size / xBarSize)
                        y_size = protein_complex[1]
                        y_size = int(y_size / yBarSize)
                        count = protein_complex[-1]

                        #Expands matrix if it is not big enough
                        if max_y_size < y_size: 
                            for i in range(max_y_size,y_size):
                                count_list.append([])
                                for i in range(max_x_size+1):
                                    count_list[-1].append(0)
                            max_y_size = y_size
                        if max_x_size < x_size: 
                            for column_index in range(len(count_list)):
                                for i in range(max_x_size,x_size):
                                    count_list[column_index].append(0)
                            max_x_size = x_size

                        #goes through each protein, than adds 1 to the main array based on # of Xs and Ys in it
                        count_list[y_size][x_size] += count

        #find max X and Y b/w all files
        if max_y_size > true_max_y:
            true_max_y = max_y_size
        if max_x_size > true_max_x:
            true_max_x = max_x_size
        

        #turns array (# of proteins with specific XY counts) into mean (b/w all timestamps) than adds it to main, multi-file list
        count_list = np.divide(count_list,data_count).tolist()
        count_list_sum.append(count_list)
    

    #makes it so the arrays all go up to max_x and max_y
    true_max_x = true_max_x + 1
    true_max_y = true_max_y + 1
    
    for file_index,array in enumerate(count_list_sum):
        y_size = len(array)
        
        if y_size < true_max_y: 
            for na1 in range(y_size,true_max_y):
                count_list_sum[file_index].append([])
                for na2 in range(true_max_x+1):
                    count_list_sum[file_index][-1].append(0)
        
        for column_index in range(len(array)):
            x_size = len(array[column_index])
            if x_size < true_max_x: 
                for i in range(x_size,true_max_x):
                    count_list_sum[file_index][column_index].append(0)
    
    #find mean and std
    count_list_mean = np.zeros([true_max_y, true_max_x])
    count_list_std = np.zeros([true_max_y, true_max_x])
    for y in range(len(count_list_sum[0])):
        for x in range(len(count_list_sum[0][0])):
            temp_list = []
            for file in count_list_sum:
                temp_list.append(file[y][x])
            count_list_mean[y][x] += np.mean(temp_list)
            count_list_std[y][x] += np.std(temp_list)
    x_list = np.arange(0, true_max_x) * xBarSize
    y_list = np.arange(0, true_max_y) * yBarSize
    
    #output variables
    if SaveVars:
        save_vars_to_file({"x_mono_count":x_list, "y_mono_count":y_list, "cmplx_count":count_list_mean, "std":count_list_std})


    #create figure
    if ShowFig:
        fig, ax = plt.subplots()
        im = ax.imshow(count_list_mean)
        ax.set_xticks(np.arange(len(x_list)))
        ax.set_yticks(np.arange(len(y_list)))
        ax.set_xticklabels(x_list)
        ax.set_yticklabels(y_list)
        if ShowMean and ShowStd:
            print('Cannot show both mean and std!')
            return 0
        if ShowMean:
            fig_name = 'Complex_Distribution_of_' + xAxis + '_and_' + yAxis + '_with_mean'
            for i in range(len(y_list)):
                for j in range(len(x_list)):
                    text = ax.text(j, i, round(
                        count_list_mean[i, j], 1), ha='center', va='center', color='w')
        elif ShowStd and FileNum != 1:
            fig_name = 'Complex_Distribution_of_' + xAxis + '_and_' + yAxis + '_with_std'
            for i in range(len(y_list)):
                for j in range(len(x_list)):
                    text = ax.text(j, i, round(
                        count_list_std[i, j], 1), ha='center', va='center', color='w')
        else:
            fig_name = 'Complex_Distribution_of_' + xAxis + '_and_' + yAxis
        ax.set_title('Complex Distribution of ' + xAxis + ' and ' + yAxis)
        fig.tight_layout()
        plt.colorbar(im)
        plt.xlabel('Count of ' + xAxis)
        plt.ylabel('Count of ' + yAxis)
        if SaveFig:
            plt.savefig(fig_name, dpi=500,  bbox_inches='tight')
        plt.show()
    return x_list, y_list, count_list_mean, count_list_std




import numpy as np


def determine_gagTemplate_structure(numGag, positionsVec):
    """
    Determine the template structure of the gags, which is the average structure of the first gag.
    The template structure is used to determine the internal basis system of each gag, and the internal
    coordinates of each interface in the internal basis system.

    Parameters:
        numGag (int): number of gags
        positionsVec (np.array): positions of all atoms in the system, rowvec
    
    Returns:
        template (np.array): the template structure of the gags, rowvec
    """

    internalBasis = np.zeros([3,3,numGag]) # 18 gags, each gag has 3 vectors of internal basises, rowvec
    coefficients = np.zeros([5,3,numGag])  # internal coords of interfaces in the internal basis system
    # set up the internal coord system for each gag: basis vec1, vec2, vec3 
    # and then calculate the coords of each interface in this internal system: internal coords
    # then the mean value of the internal coords gives the template structure 
    for i in range (0,numGag):
        center = positionsVec[6*i,:]
        interfaces = positionsVec[1+6*i:5+1+6*i,:]
        # determine the three internal basis: vec1, vec2, vec3
        vec1 = center
        vec1 = vec1/np.linalg.norm(vec1) 
        vec2 = interfaces[0,:] - center
        vec3 = np.cross(vec1,vec2)
        vec3 = vec3/np.linalg.norm(vec3)
        vec2 = np.cross(vec3,vec1)
        vec2 = vec2/np.linalg.norm(vec2)
        internalBasis[0,:,i] = vec1
        internalBasis[1,:,i] = vec2
        internalBasis[2,:,i] = vec3
        # calculate the interal coords for the 5 interfaces
        for j in range (0,5):
            p = interfaces[j,:] - center
            A = np.array([vec1, vec2, vec3])
            coeff = np.dot(p, np.linalg.inv(A))
            if (np.linalg.norm(np.dot(coeff,A)-p) > 1e-12) : # check whether correctly calculated!
                print('Wrong calculation of coefficients\n',np.linalg.norm(np.dot(coeff,A)-p))
            coefficients[j,:,i] = coeff
    
    # regularize the gags internal coords
    coeffReg = np.zeros([5,3]) # five sites, each site has 3 coefficients of internal coords
    for i in range(0,5) :
        coeffReg[i,0] = np.mean(coefficients[i,0,:])
        coeffReg[i,1] = np.mean(coefficients[i,1,:])
        coeffReg[i,2] = np.mean(coefficients[i,2,:])

    # using the mean coefficients to calculate the structure of the first gag, and take it as the template
    chosenGagIndex = 0
    vec1 = internalBasis[0,:,chosenGagIndex]
    vec2 = internalBasis[1,:,chosenGagIndex]
    vec3 = internalBasis[2,:,chosenGagIndex]
    center1 = positionsVec[0+6*chosenGagIndex,:]
    template = np.zeros([6,3])
    template[0,:] = center1
    for i in range(0,5):
        template[i+1,:] = coeffReg[i,0] * vec1 + coeffReg[i,1] * vec2 + coeffReg[i,2] * vec3 + center1
    
    return template




def write_new_PDB(protein_remain, main_pdb_list, OpName):
    """Generates a new PDB file with protein information based on a list of remaining protein numbers.

    This function reads a PDB file, extracts protein information for the proteins whose numbers are specified
    in the `protein_remain` list, and writes the extracted information to a new PDB file.

    Args:
        protein_remain (list): A list of protein numbers that had the correct number of sub-proteins.
        main_pdb_list (list): A list of lists of every single line of the original inputted pdb file. info[i][0] = important data from line, info[i][1] = the line string
        OpName (str): The name of the outputted file. 

    Returns:
        .pdb file: holds all of the proteins sites that are in the complexes of the correct sizes
    """

    OpName = f"{OpName}.pdb"
    
    #write new file
    with open(OpName, 'w') as file:
        file.seek(0)
        file.truncate()
        
        #goes through the main_PDB_list which has a setup of [['info','string of line'],[],[],...] with info being the important data in that line (or description of line)
        for line in main_pdb_list:
            if line[0] == "header":
                file.writelines(line[1])
            elif line[0] in protein_remain:
                file.writelines(line[1])
    return 0





from .read_PDB import read_PDB
from .read_inp import read_inp
from .create_bond_list import create_bond_list
from .create_complex_list import create_complex_list
from .filter_complexes import filter_complexes
from .write_new_PDB import write_new_PDB


def locate_pos_no_restart(FileNamePdb, NumDict, FileNameInp, BufferRatio=0.01, OpName = "output_file"):
    """
    Locates specific complexes of a certain size from a PDB file after simulation and outputs the result as a separated file
    named "output_file.pdb" containing only the desired complex.

    Args:
        FileNamePdb (str): The path to the PDB file, which is usually the last frame of the simulation.
        NumDict (dictionary): A dictionary that holds the requested number of protein types in a complex
            Ex: {'dod':11,'clat':4}
        FileNameInp (str): The path to the '.inp' file, which usually stores the reaction information.
        BufferRatio (float, optional): The buffer ratio used to determine whether two reaction interfaces can be considered as bonded.
            Defaults to 0.01.
        OpName (str, optional = “output_file”): The name of the outputted file. 

    Returns:
        output_file.pdb: A file containing the desired complex.

    Note:
        This function is slightly slower than the one that reads restart, but still runs quite fast.

    Examples:
        >>> locate_position_PDB('/Users/UserName/Documents/999999.pdb', [12], '/Users/UserName/Documents/parms.inp', 0.05)
        "/Users/UserName/Documents/output_file.pdb"
    """

    #reads in the .pdb file
    print('Reading files......')
    site_array,site_dict,num_name_dict,main_pdb_list = read_PDB(FileNamePdb, True)
    print('Reading files complete!')

    #reads in the .inp file
    print('Extracting binding information......')
    binding_array,binding_dict = read_inp(FileNameInp)
    print('Extracting complete!')

    #creates list of every bond
    print('Calculating distance......')
    bonds_lst = create_bond_list(site_array,site_dict,binding_array,binding_dict,BufferRatio)
    print('Calculation complete!')

    #creates list of each complex
    print('Finding complexes......')
    complex_lst = create_complex_list(bonds_lst)
    print('Finding complexes complete!')

    #creates list of each complex that has the correct number of each type
    print('Filtering complexes......')
    complex_filtered = filter_complexes(complex_lst,num_name_dict,NumDict)
    print('Filtering complexes complete!')

    #writes a new PDB file that only includs proteins that are in complexes of the correct size
    print('Writing new PDB files......')
    write_new_PDB(complex_filtered, main_pdb_list, OpName)
    print(f'PDB writing complete! (named as {OpName}.pdb)')
    return 0





def octa_face_vert_coord(radius: float):
    """Generates the coordinates of the vertices of an octahedron based on the given radius.

    Args:
        radius (float): The radius of the octahedron.

    Returns:
        List[List[float]]: A list of vertex coordinates of the octahedron. Each element in the list
            is a sublist containing three floating point values representing the (x, y, z) coordinates
            of a vertex.

    Example:
        radius = 5.0
        vert_coord = octa_face_vert_coord(radius)
        print(vert_coord)

    Note:
        The function generates the vertex coordinates of an octahedron centered at the origin (0, 0, 0)
        with six vertices located at (+-radius, 0, 0), (0, +-radius, 0), and (0, 0, +-radius). The
        resulting vertex coordinates are returned as a list of lists, where each sublist contains the
        (x, y, z) coordinates of a specific vertex.
    """
    scaler = radius
    v0 = [1, 0, 0]
    v1 = [-1, 0, 0]
    v2 = [0, 1, 0]
    v3 = [0, -1, 0]
    v4 = [0, 0, 1]
    v5 = [0, 0, -1]
    VertCoord = [v0, v1, v2, v3, v4, v5]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





def cube_face_vert_coord(radius: float):
    """Generates vertex coordinates for a cube face.

    This function takes the radius of a cube face as input, and generates a list of vertex coordinates for the cube face
    of a platonic solid. The vertex coordinates are calculated by scaling the pre-defined vertex coordinates of a unit cube
    by the given radius value.

    Args:
        radius (float): The radius of the platonic solid.

    Returns:
        List: Contains vertex coordinates for the cube face. Each vertex coordinate is a list of three floats representing the
        x, y, and z coordinates of the vertex. The vertex coordinates are scaled by the radius value. 

    Raises:
        None.

    Example:
        >>> cube_face_vert_coord(1.0)
        [[0.5773502691896257, 0.5773502691896257, 0.5773502691896257],
         [-0.5773502691896257, 0.5773502691896257, 0.5773502691896257],
         [0.5773502691896257, -0.5773502691896257, 0.5773502691896257],
         [0.5773502691896257, 0.5773502691896257, -0.5773502691896257],
         [-0.5773502691896257, -0.5773502691896257, 0.5773502691896257],
         [0.5773502691896257, -0.5773502691896257, -0.5773502691896257],
         [-0.5773502691896257, 0.5773502691896257, -0.5773502691896257],
         [-0.5773502691896257, -0.5773502691896257, -0.5773502691896257]]
    """
    
    scaler = radius/3**0.5
    v0 = [1, 1, 1]
    v1 = [-1, 1, 1]
    v2 = [1, -1, 1]
    v3 = [1, 1, -1]
    v4 = [-1, -1, 1]
    v5 = [1, -1, -1]
    v6 = [-1, 1, -1]
    v7 = [-1, -1, -1]
    VertCoord = [v0, v1, v2, v3, v4, v5, v6, v7]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





from .octa_vert_write import octa_vert_write


def octa_vert(radius: float, sigma: float):
    """
    Writes octagonal vertices to a file.

    Args:
        radius (float): The radius of the octagon.
        sigma (float): The standard deviation for the Gaussian distribution.

    Returns:
        parm.inp/icos.mol: input files for NERDSS

    Example:
        octa_vert(5.0, 1.0)
    """
    octa_vert_write(radius, sigma)
    print('File writing complete!')
    return 0





import math
from ..gen_platonic.distance import distance


def tetr_face_leg_reduce(COM: float, leg: float, sigma: float):
    """Reduces the length of a leg of a tetrahedron face given its center of mass (COM), the original length of the leg,
    and a scaling factor sigma.

    Args:
        COM (float): The coordinates of the center of mass of the tetrahedron face as a list of 3 floats representing
        the x, y, and z coordinates.
        leg (float): The coordinates of the original leg of the tetrahedron face as a list of 3 floats representing
        the x, y, and z coordinates.
        sigma (float): A scaling factor for reducing the length of the leg.

    Returns:
        list: A list of 3 floats representing the reduced coordinates of the leg after applying the scaling factor.

    Example:
        >>> tetr_face_leg_reduce([0.0, 0.0, 0.0], [-0.5, 0.0, 0.0], 0.5)
        [-0.25, 0.0, 0.0]
    """
    n = 12
    angle = math.acos(1/3)
    red_len = sigma/(2*math.sin(angle/2))
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], n))
    return leg_red


